# stdlib
from string import Template
from typing import Any

# third party
from pydantic import ValidationError

# relative
from ...abstract_server import ServerSideType
from ...serde.serializable import serializable
from ...store.db.db import DBManager
from ...store.document_store_errors import NotFoundException
from ...store.document_store_errors import StashException
from ...types.errors import SyftException
from ...types.result import as_result
from ...types.syft_metaclass import Empty
from ...util.assets import load_png_base64
from ...util.experimental_flags import flags
from ...util.misc_objs import HTMLObject
from ...util.misc_objs import MarkdownDescription
from ...util.notebook_ui.styles import FONT_CSS
from ...util.schema import DO_COMMANDS
from ...util.schema import DS_COMMANDS
from ...util.schema import GUEST_COMMANDS
from ..context import AuthedServiceContext
from ..context import UnauthedServiceContext
from ..notifier.notifier_enums import EMAIL_TYPES
from ..notifier.notifier_enums import NOTIFICATION_FREQUENCY
from ..response import SyftSuccess
from ..service import AbstractService
from ..service import service_method
from ..user.user_roles import ADMIN_ROLE_LEVEL
from ..user.user_roles import GUEST_ROLE_LEVEL
from ..user.user_roles import ServiceRole
from ..warnings import HighSideCRUDWarning
from .settings import ServerSettings
from .settings import ServerSettingsUpdate
from .settings_stash import SettingsStash

# for testing purpose
_NOTIFICATIONS_ENABLED_WIHOUT_CREDENTIALS_ERROR = (
    "Failed to enable notification. "
    "Email credentials are invalid or have not been set. "
    "Please use `enable_notifications` from `user_service` "
    "to set the correct email credentials."
)


@serializable(canonical_name="SettingsService", version=1)
class SettingsService(AbstractService):
    stash: SettingsStash

    def __init__(self, store: DBManager) -> None:
        self.stash = SettingsStash(store=store)

    @service_method(path="settings.get", name="get")
    def get(self, context: UnauthedServiceContext) -> ServerSettings:
        """Get Settings"""
        all_settings = self.stash.get_all(
            context.server.signing_key.verify_key
        ).unwrap()

        if len(all_settings) == 0:
            raise NotFoundException(public_message="No settings found")

        return all_settings[0]

    @service_method(path="settings.set", name="set")
    def set(
        self, context: AuthedServiceContext, settings: ServerSettings
    ) -> ServerSettings:
        """Set a new the Server Settings"""
        return self.stash.set(context.credentials, settings).unwrap()

    @service_method(
        path="settings.update",
        name="update",
        autosplat=["settings"],
        unwrap_on_success=False,
        roles=ADMIN_ROLE_LEVEL,
    )
    def update(
        self, context: AuthedServiceContext, settings: ServerSettingsUpdate
    ) -> SyftSuccess:
        """
        Update the Server Settings using the provided values.

        Args:
            name: Optional[str]
                Server name
            organization: Optional[str]
                Organization name
            description: Optional[str]
                Server description
            on_board: Optional[bool]
                Show onboarding panel when a user logs in for the first time
            signup_enabled: Optional[bool]
                Enable/Disable registration
            admin_email: Optional[str]
                Administrator email
            association_request_auto_approval: Optional[bool]

        Returns:
            SyftSuccess: Message indicating the success of the operation, with the
                update server settings as the value property.

        Example:
        >>> server_client.update(name='foo', organization='bar', description='baz', signup_enabled=True)
        SyftSuccess: Settings updated successfully.
        """
        updated_settings = self._update(context, settings).unwrap()
        return SyftSuccess(
            message=(
                "Settings updated successfully. "
                + "You must call <client>.refresh() to sync your client with the changes."
            ),
            value=updated_settings,
        )

    @as_result(StashException, NotFoundException, ValidationError)
    def _update(
        self, context: AuthedServiceContext, settings: ServerSettingsUpdate
    ) -> ServerSettings:
        all_settings = self.stash.get_all(
            context.credentials, limit=1, sort_order="desc"
        ).unwrap()
        if len(all_settings) > 0:
            new_settings = all_settings[0].model_copy(
                update=settings.to_dict(exclude_empty=True)
            )
            ServerSettings.model_validate(new_settings.to_dict())
            update_result = self.stash.update(
                context.credentials, obj=new_settings
            ).unwrap()

            # If notifications_enabled is present in the update, we need to update the notifier settings
            if settings.notifications_enabled is not Empty:  # type: ignore[comparison-overlap]
                notifier_settings_res = context.server.services.notifier.settings(
                    context
                )
                if (
                    not notifier_settings_res.is_ok()
                    or notifier_settings_res.ok() is None
                ):
                    raise SyftException(
                        public_message=(
                            "Notification has not been enabled. "
                            "Please use `enable_notifications` from `user_service`."
                        )
                    )

                context.server.services.notifier._set_notifier(
                    context, active=settings.notifications_enabled
                )

            return update_result
        else:
            raise NotFoundException(public_message="Server settings not found")

    @service_method(
        path="settings.set_server_side_type_dangerous",
        name="set_server_side_type_dangerous",
        roles=ADMIN_ROLE_LEVEL,
        unwrap_on_success=False,
    )
    def set_server_side_type_dangerous(
        self, context: AuthedServiceContext, server_side_type: str
    ) -> SyftSuccess:
        side_type_options = [e.value for e in ServerSideType]

        if server_side_type not in side_type_options:
            raise SyftException(
                public_message=f"Not a valid server_side_type, please use one of the options from: {side_type_options}"
            )

        current_settings = self.stash.get_all(
            context.credentials, limit=1, sort_order="desc"
        ).unwrap()
        if len(current_settings) > 0:
            new_settings = current_settings[0]
            new_settings.server_side_type = ServerSideType(server_side_type)
            updated_settings = self.stash.update(
                context.credentials, new_settings
            ).unwrap()
            return SyftSuccess(
                message=(
                    "Settings updated successfully. "
                    + "You must call <client>.refresh() to sync your client with the changes."
                ),
                value=updated_settings,
            )
        else:
            # TODO: Turn this into a function?
            raise NotFoundException(public_message="Server settings not found")

    @service_method(
        path="settings.batch_notifications",
        name="batch_notifications",
        roles=ADMIN_ROLE_LEVEL,
    )
    def batch_notifications(
        self,
        context: AuthedServiceContext,
        email_type: EMAIL_TYPES,
        frequency: NOTIFICATION_FREQUENCY,
        start_time: str = "",
    ) -> SyftSuccess:
        result = context.server.services.notifier.set_email_batch(
            context=context,
            email_type=email_type,
            frequency=frequency,
            start_time=start_time,
        ).unwrap()
        return result

    @service_method(
        path="settings.enable_notifications",
        name="enable_notifications",
        roles=ADMIN_ROLE_LEVEL,
    )
    def enable_notifications(
        self,
        context: AuthedServiceContext,
        email_username: str | None = None,
        email_password: str | None = None,
        email_sender: str | None = None,
        email_server: str | None = None,
        email_port: str | None = None,
    ) -> SyftSuccess:
        context.server.services.notifier.turn_on(
            context=context,
            email_username=email_username,
            email_password=email_password,
            email_sender=email_sender,
            email_server=email_server,
            email_port=email_port,
        ).unwrap()
        return SyftSuccess(message="Notifications enabled")

    @service_method(
        path="settings.disable_notifications",
        name="disable_notifications",
        roles=ADMIN_ROLE_LEVEL,
    )
    def disable_notifications(
        self,
        context: AuthedServiceContext,
    ) -> SyftSuccess:
        context.server.services.notifier.turn_off(context=context).unwrap()
        return SyftSuccess(message="Notifications disabled")

    @service_method(
        path="settings.allow_guest_signup",
        name="allow_guest_signup",
        warning=HighSideCRUDWarning(confirmation=True),
        unwrap_on_success=False,
    )
    def allow_guest_signup(
        self, context: AuthedServiceContext, enable: bool
    ) -> SyftSuccess:
        """Enable/Disable Registration for Data Scientist or Guest Users."""
        flags.CAN_REGISTER = enable

        settings = ServerSettingsUpdate(signup_enabled=enable)
        self._update(context=context, settings=settings).unwrap()
        message = "enabled" if enable else "disabled"
        return SyftSuccess(
            message=f"Registration feature successfully {message}", value=message
        )

    # NOTE: This service is disabled until we bring back Eager Execution
    # @service_method(
    #     path="settings.enable_eager_execution",
    #     name="enable_eager_execution",
    #     roles=ADMIN_ROLE_LEVEL,
    #     warning=HighSideCRUDWarning(confirmation=True),
    # )
    def enable_eager_execution(
        self, context: AuthedServiceContext, enable: bool
    ) -> SyftSuccess:
        """Enable/Disable eager execution."""
        settings = ServerSettingsUpdate(eager_execution_enabled=enable)
        self._update(context=context, settings=settings).unwrap()
        message = "enabled" if enable else "disabled"
        return SyftSuccess(message=f"Eager execution {message}", value=message)

    @service_method(path="settings.set_email_rate_limit", name="set_email_rate_limit")
    def set_email_rate_limit(
        self, context: AuthedServiceContext, email_type: EMAIL_TYPES, daily_limit: int
    ) -> SyftSuccess:
        return context.server.services.notifier.set_email_rate_limit(
            context, email_type, daily_limit
        )

    @service_method(
        path="settings.allow_association_request_auto_approval",
        name="allow_association_request_auto_approval",
        unwrap_on_success=False,
    )
    def allow_association_request_auto_approval(
        self, context: AuthedServiceContext, enable: bool
    ) -> SyftSuccess:
        new_settings = ServerSettingsUpdate(association_request_auto_approval=enable)
        self._update(context, settings=new_settings).unwrap()
        message = "enabled" if enable else "disabled"
        return SyftSuccess(
            message="Association request auto-approval successfully " + message
        )

    @service_method(
        path="settings.welcome_preview",
        name="welcome_preview",
    )
    def welcome_preview(
        self,
        context: AuthedServiceContext,
        markdown: str = "",
        html: str = "",
    ) -> MarkdownDescription | HTMLObject:
        if not markdown and not html or markdown and html:
            raise SyftException(
                public_message="Invalid markdown/html fields. You must set one of them."
            )

        welcome_msg = None
        if markdown:
            welcome_msg = MarkdownDescription(text=markdown)
        else:
            welcome_msg = HTMLObject(text=html)

        return welcome_msg

    @service_method(
        path="settings.welcome_customize",
        name="welcome_customize",
        unwrap_on_success=False,
    )
    def welcome_customize(
        self,
        context: AuthedServiceContext,
        markdown: str = "",
        html: str = "",
    ) -> SyftSuccess:
        if not markdown and not html or markdown and html:
            raise SyftException(
                public_message="Invalid markdown/html fields. You must set one of them."
            )

        welcome_msg = None
        if markdown:
            welcome_msg = MarkdownDescription(text=markdown)
        else:
            welcome_msg = HTMLObject(text=html)

        new_settings = ServerSettingsUpdate(welcome_markdown=welcome_msg)
        self._update(context=context, settings=new_settings).unwrap()

        return SyftSuccess(message="Welcome Markdown was successfully updated!")

    @service_method(
        path="settings.welcome_show",
        name="welcome_show",
        roles=GUEST_ROLE_LEVEL,
    )
    def welcome_show(
        self,
        context: AuthedServiceContext,
    ) -> HTMLObject | MarkdownDescription:
        all_settings = self.stash.get_all(
            context.server.signing_key.verify_key
        ).unwrap()
        role = context.server.services.user.get_role_for_credentials(
            context.credentials
        ).unwrap()

        # check if the settings list is empty
        if len(all_settings) == 0:
            raise NotFoundException(public_message="Server settings not found")
        settings = all_settings[0]

        if settings.welcome_markdown:
            str_tmp = Template(settings.welcome_markdown.text)
            welcome_msg_class = type(settings.welcome_markdown)
            server_side_type = (
                "Low Side"
                if context.server.metadata.server_side_type
                == ServerSideType.LOW_SIDE.value
                else "High Side"
            )
            commands = ""
            if (
                role.value == ServiceRole.NONE.value
                or role.value == ServiceRole.GUEST.value
            ):
                commands = GUEST_COMMANDS
            elif role is not None and role.value == ServiceRole.DATA_SCIENTIST.value:
                commands = DS_COMMANDS
            elif role is not None and role.value >= ServiceRole.DATA_OWNER.value:
                commands = DO_COMMANDS

            command_list = f"""
            <ul style='padding-left: 1em;'>
                {commands}
            </ul>
            """
            result = str_tmp.safe_substitute(
                FONT_CSS=FONT_CSS,
                server_symbol=load_png_base64("small-syft-symbol-logo.png"),
                datasite_name=context.server.name,
                description=context.server.metadata.description,
                # server_url='http://testing:8080',
                server_type=context.server.metadata.server_type.capitalize(),
                server_side_type=server_side_type,
                server_version=context.server.metadata.syft_version,
                command_list=command_list,
            )
            return welcome_msg_class(text=result)
        raise SyftException(public_message="There's no welcome message")

    @service_method(
        path="settings.get_server_config",
        name="get_server_config",
        roles=ADMIN_ROLE_LEVEL,
    )
    def get_server_config(
        self,
        context: AuthedServiceContext,
    ) -> dict[str, Any]:
        server = context.server

        return {
            "name": server.name,
            "server_type": server.server_type,
            # "deploy_to": server.deployment_type_enum,
            "server_side_type": server.server_side_type,
            # "port": server.port,
            "processes": server.processes,
            "dev_mode": server.dev_mode,
            "reset": True,  # we should be able to get all the objects from migration data
            "tail": False,
            # "host": server.host,
            "enable_warnings": server.enable_warnings,
            "n_consumers": server.queue_config.client_config.create_producer,
            "thread_workers": server.queue_config.thread_workers,
            "create_producer": server.queue_config.client_config.create_producer,
            "queue_port": server.queue_config.client_config.queue_port,
            "association_request_auto_approval": server.association_request_auto_approval,
            "background_tasks": True,
            "debug": True,  # we also want to debug
            "migrate": False,  # I think we dont want to migrate?
        }



# stdlib

# relative
from ...serde.serializable import serializable
from ...store.db.db import DBManager
from ..service import AbstractService


@serializable(canonical_name="EnclaveService", version=1)
class EnclaveService(AbstractService):
    def __init__(self, store: DBManager) -> None:
        pass



# stdlib
import os

# relative
from ...util import str_to_bool
from .submit_query import make_submit_query

env_var = "TEST_BIGQUERY_APIS_LIVE"
use_live = str_to_bool(str(os.environ.get(env_var, "False")))
env_name = "Live" if use_live else "Mock"
print(f"Using {env_name} API Code, this will query BigQuery. ${env_var}=={use_live}")


if use_live:
    # relative
    from .live.schema import make_schema
    from .live.test_query import make_test_query
else:
    # relative
    from .mock.schema import make_schema
    from .mock.test_query import make_test_query



# third party
from pydantic import Field

# relative
from ..serde.serializable import serializable
from ..types.syft_object import StorableObjectType
from .document_store import DocumentStore
from .document_store import StoreConfig
from .document_store import StorePartition
from .kv_document_store import KeyValueBackingStore
from .locks import LockingConfig
from .locks import NoLockingConfig
from .mongo_client import MongoStoreClientConfig


class MongoBsonObject(StorableObjectType):
    pass


@serializable(
    attrs=["index_name", "settings", "store_config"],
    canonical_name="MongoBackingStore",
    version=1,
)
class MongoBackingStore(KeyValueBackingStore):
    pass


@serializable(attrs=["storage_type"], canonical_name="MongoStorePartition", version=1)
class MongoStorePartition(StorePartition):
    """Mongo StorePartition
    Parameters:
        `settings`: PartitionSettings
            PySyft specific settings, used for partitioning and indexing.
        `store_config`: MongoStoreConfig
            Mongo specific configuration
    """

    storage_type: type[StorableObjectType] = MongoBsonObject


@serializable(canonical_name="MongoDocumentStore", version=1)
class MongoDocumentStore(DocumentStore):
    """Mongo Document Store
    Parameters:
        `store_config`: MongoStoreConfig
            Mongo specific configuration, including connection configuration, database name, or client class type.
    """

    partition_type = MongoStorePartition


@serializable()
class MongoStoreConfig(StoreConfig):
    __canonical_name__ = "MongoStoreConfig"

    """Mongo Store configuration
    Parameters:
        `client_config`: MongoStoreClientConfig
            Mongo connection details: hostname, port, user, password etc.
        `store_type`: Type[DocumentStore]
            The type of the DocumentStore. Default: MongoDocumentStore
        `db_name`: str
            Database name
        locking_config: LockingConfig
            The config used for store locking. Available options:
                * NoLockingConfig: no locking, ideal for single-thread stores.
                * ThreadingLockingConfig: threading-based locking, ideal for same-process in-memory stores.
            Defaults to NoLockingConfig.
    """

    client_config: MongoStoreClientConfig
    store_type: type[DocumentStore] = MongoDocumentStore
    db_name: str = "app"
    backing_store: type[KeyValueBackingStore] = MongoBackingStore
    # TODO: should use a distributed lock, with RedisLockingConfig
    locking_config: LockingConfig = Field(default_factory=NoLockingConfig)



# stdlib
import builtins as __builtin__
import datetime
import sys
from typing import Any

# relative
from ...types.uid import UID
from ..action.action_object import ActionObject
from ..context import AuthedServiceContext
from ..job.job_stash import Job
from ..response import SyftError


def print(
    context: AuthedServiceContext,
    log_id: UID,
    *args: Any,
    sep: str = " ",
    end: str = "\n",
) -> str | None:
    def to_str(arg: Any) -> str:
        if isinstance(arg, bytes):
            return arg.decode("utf-8")
        if isinstance(arg, Job):
            return f"JOB: {arg.id}"
        if isinstance(arg, SyftError):
            return f"JOB: {arg.message}"
        if isinstance(arg, ActionObject):
            return str(arg.syft_action_data)
        return str(arg)

    new_args = [to_str(arg) for arg in args]
    new_str = sep.join(new_args) + end
    if context.server is not None:
        context.server.services.log.append(context=context, uid=log_id, new_str=new_str)
    time = datetime.datetime.now().strftime("%d/%m/%y %H:%M:%S")
    return __builtin__.print(
        f"{time} FUNCTION LOG :",
        *new_args,
        end=end,
        sep=sep,
        file=sys.stderr,
    )



# stdlib
from collections.abc import Collection
from collections.abc import Sequence
import logging

# relative
from ...serde.serializable import serializable
from ...store.db.db import DBManager
from ...types.dicttuple import DictTuple
from ...types.uid import UID
from ..action.action_permissions import ActionObjectPermission
from ..action.action_permissions import ActionPermission
from ..context import AuthedServiceContext
from ..response import SyftSuccess
from ..service import AbstractService
from ..service import SERVICE_TO_TYPES
from ..service import TYPE_TO_SERVICE
from ..service import service_method
from ..user.user_roles import DATA_OWNER_ROLE_LEVEL
from ..user.user_roles import DATA_SCIENTIST_ROLE_LEVEL
from ..user.user_roles import GUEST_ROLE_LEVEL
from ..warnings import CRUDReminder
from ..warnings import HighSideCRUDWarning
from .dataset import Asset
from .dataset import CreateDataset
from .dataset import Dataset
from .dataset import DatasetPageView
from .dataset_stash import DatasetStash

logger = logging.getLogger(__name__)


def _paginate_collection(
    collection: Collection,
    page_size: int | None = 0,
    page_index: int | None = 0,
) -> slice | None:
    if page_size is None or page_size <= 0:
        return None

    # If chunk size is defined, then split list into evenly sized chunks
    total = len(collection)
    page_index = 0 if page_index is None else page_index

    if page_size > total or page_index >= total // page_size or page_index < 0:
        return None

    start = page_size * page_index
    stop = min(page_size * (page_index + 1), total)
    return slice(start, stop)


def _paginate_dataset_collection(
    datasets: Sequence[Dataset],
    page_size: int | None = 0,
    page_index: int | None = 0,
) -> DictTuple[str, Dataset] | DatasetPageView:
    slice_ = _paginate_collection(datasets, page_size=page_size, page_index=page_index)
    chunk = datasets[slice_] if slice_ is not None else datasets
    results = DictTuple(chunk, lambda dataset: dataset.name)

    return (
        results
        if slice_ is None
        else DatasetPageView(datasets=results, total=len(datasets))
    )


@serializable(canonical_name="DatasetService", version=1)
class DatasetService(AbstractService):
    stash: DatasetStash

    def __init__(self, store: DBManager) -> None:
        self.stash = DatasetStash(store=store)

    @service_method(
        path="dataset.add",
        name="add",
        roles=DATA_OWNER_ROLE_LEVEL,
        unwrap_on_success=False,
    )
    def add(self, context: AuthedServiceContext, dataset: CreateDataset) -> SyftSuccess:
        """Add a Dataset"""
        dataset = dataset.to(Dataset, context=context)

        result = self.stash.set(
            context.credentials,
            dataset,
            add_permissions=[
                ActionObjectPermission(
                    uid=dataset.id, permission=ActionPermission.ALL_READ
                ),
            ],
        ).unwrap()

        return SyftSuccess(
            message=(
                f"Dataset uploaded to '{context.server.name}'."
                f" To see the datasets uploaded by a client on this server, use command `[your_client].datasets`"
            ),
            value=result,
        )

    @service_method(
        path="dataset.get_all",
        name="get_all",
        roles=GUEST_ROLE_LEVEL,
        warning=CRUDReminder(),
    )
    def get_all(
        self,
        context: AuthedServiceContext,
        page_size: int | None = 0,
        page_index: int | None = 0,
    ) -> DatasetPageView | DictTuple[str, Dataset]:
        """Get a Dataset"""
        datasets = self.stash.get_all_active(context.credentials).unwrap()

        for dataset in datasets:
            if context.server is not None:
                dataset.server_uid = context.server.id

        return _paginate_dataset_collection(
            datasets=datasets, page_size=page_size, page_index=page_index
        )

    @service_method(path="dataset.search", name="search", roles=GUEST_ROLE_LEVEL)
    def search(
        self,
        context: AuthedServiceContext,
        name: str,
        page_size: int | None = 0,
        page_index: int | None = 0,
    ) -> DatasetPageView | DictTuple[str, Dataset]:
        """Search a Dataset by name"""
        results = self.get_all(context)

        filtered_results = [
            dataset
            for dataset_name, dataset in results.items()
            if name in dataset_name and not dataset.to_be_deleted
        ]

        return _paginate_dataset_collection(
            filtered_results, page_size=page_size, page_index=page_index
        )

    @service_method(path="dataset.get_by_id", name="get_by_id")
    def get_by_id(self, context: AuthedServiceContext, uid: UID) -> Dataset:
        """Get a Dataset"""
        dataset = self.stash.get_by_uid(context.credentials, uid=uid).unwrap()

        if context.server is not None:
            dataset.server_uid = context.server.id

        return dataset

    @service_method(path="dataset.get_by_action_id", name="get_by_action_id")
    def get_by_action_id(
        self, context: AuthedServiceContext, uid: UID
    ) -> list[Dataset]:
        """Get Datasets by an Action ID"""
        datasets = self.stash.search_action_ids(context.credentials, uid=uid).unwrap()

        for dataset in datasets:
            if context.server is not None:
                dataset.server_uid = context.server.id

        return datasets

    @service_method(
        path="dataset.get_assets_by_action_id",
        name="get_assets_by_action_id",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def get_assets_by_action_id(
        self, context: AuthedServiceContext, uid: UID
    ) -> list[Asset]:
        """Get Assets by an Action ID"""
        datasets = self.get_by_action_id(context=context, uid=uid)

        for dataset in datasets:
            if dataset.to_be_deleted:
                datasets.remove(dataset)

        return [
            asset
            for dataset in datasets
            for asset in dataset.asset_list
            if asset.action_id == uid
        ]

    @service_method(
        path="dataset.delete",
        name="delete",
        roles=DATA_OWNER_ROLE_LEVEL,
        warning=HighSideCRUDWarning(confirmation=True),
        unwrap_on_success=False,
    )
    def delete(
        self, context: AuthedServiceContext, uid: UID, delete_assets: bool = True
    ) -> SyftSuccess:
        """
        Soft delete: keep the dataset object, only remove the blob store entries
        After soft deleting a dataset, the user will not be able to
        see it using the `datasets.get_all` endpoint.
        Delete unique `dataset.name` key and leave UID, just rename it in case the
        user upload a new dataset with the same name.
        """
        # check if the dataset exists
        dataset = self.get_by_id(context=context, uid=uid)

        return_msg = []

        if delete_assets:
            # delete the dataset's assets
            for asset in dataset.asset_list:
                msg = (
                    f"ActionObject {asset.action_id} "
                    f"linked with Assset {asset.id} "
                    f"in Dataset {uid}"
                )

                context.server.services.action.delete(
                    context=context, uid=asset.action_id, soft_delete=True
                )

                logger.info(f"Successfully deleted {msg}")
                return_msg.append(f"Asset with id '{asset.id}' successfully deleted.")

        # soft delete the dataset object from the store
        dataset.name = f"_deleted_{dataset.name}_{uid}"
        dataset.to_be_deleted = True
        self.stash.update(context.credentials, dataset).unwrap()
        return_msg.append(f"Dataset with id '{uid}' successfully deleted.")
        return SyftSuccess(message="\n".join(return_msg))


TYPE_TO_SERVICE[Dataset] = DatasetService
SERVICE_TO_TYPES[DatasetService].update({Dataset})



# stdlib
import os
from pathlib import Path
import tempfile
import zipfile

# syft absolute
from syft import SyftError
from syft import SyftException
from syft.client.client import SyftClient
from syft.service.user.user_roles import ServiceRole
from syft.util.util import get_root_data_path

# relative
from ...server.env import get_default_root_email
from ...server.env import get_default_root_password
from .worker_helpers import build_and_push_image

CHECKPOINT_ROOT = "checkpoints"
CHECKPOINT_DIR_PREFIX = "chkpt"
DEFAULT_CHECKPOINT_DIR = get_root_data_path() / CHECKPOINT_ROOT

try:
    # Ensure the default checkpoint path exists always
    DEFAULT_CHECKPOINT_DIR.mkdir(parents=True, exist_ok=True)
except Exception as e:
    print(f"Error creating default checkpoint directory: {e}")


def is_admin(client: SyftClient) -> bool:
    return client._SyftClient__user_role == ServiceRole.ADMIN


def is_valid_dir(path: Path | str) -> Path:
    if isinstance(path, str):
        path = Path(path)
    if not path.is_dir():
        raise SyftException(f"Path {path} is not a directory.")
    return path


def is_valid_file(path: Path | str) -> Path:
    if isinstance(path, str):
        path = Path(path)
    if not path.is_file():
        raise SyftException(f"Path {path} is not a file.")
    return path


def create_checkpoint(
    name: str,  # Name of the checkpoint
    client: SyftClient,
    chkpt_dir: Path | str = DEFAULT_CHECKPOINT_DIR,
    root_email: str | None = None,
    root_pwd: str | None = None,
) -> None:
    """Save a checkpoint for the database."""

    is_valid_dir(chkpt_dir)

    if root_email is None:
        root_email = get_default_root_email()

    if root_pwd is None:
        root_pwd = get_default_root_password()

    root_client = (
        client
        if is_admin(client)
        else client.login(email=root_email, password=root_pwd)
    )
    migration_data = root_client.get_migration_data(include_blobs=True)

    if isinstance(migration_data, SyftError):
        raise SyftException(message=migration_data.message)

    checkpoint_path = chkpt_dir / f"{name}.zip"

    # get a temporary directory to save the checkpoint
    temp_dir = Path(tempfile.mkdtemp())
    checkpoint_blob = temp_dir / "checkpoint.blob"
    checkpoint_yaml = temp_dir / "checkpoint.yaml"
    migration_data.save(
        path=checkpoint_blob,
        yaml_path=checkpoint_yaml,
    )

    # Combine the files into a single zip file to checkpoint_path
    with zipfile.ZipFile(checkpoint_path, "w") as zipf:
        zipf.write(checkpoint_blob, "checkpoint.blob")
        zipf.write(checkpoint_yaml, "checkpoint.yaml")

    print(f"Checkpoint saved at: \n {checkpoint_path}")


def get_checkpoint_for(
    path: Path | str | None = None, chkpt_name: str | None = None
) -> Path | None:
    # Path takes precedence over name
    if path:
        return is_valid_file(path)

    if chkpt_name:
        return is_valid_file(DEFAULT_CHECKPOINT_DIR / f"{chkpt_name}.zip")


def get_registry_credentials() -> tuple[str, str]:
    return os.environ.get("REGISTRY_USERNAME", ""), os.environ.get(
        "REGISTRY_PASSWORD", ""
    )


def load_from_checkpoint(
    client: SyftClient,
    name: str | None = None,
    path: Path | str | None = None,
    root_email: str | None = None,
    root_password: str | None = None,
    registry_username: str | None = None,
    registry_password: str | None = None,
) -> None:
    """Load the last saved checkpoint for the given checkpoint state."""

    root_email = "info@openmined.org" if root_email is None else root_email
    root_password = "changethis" if root_password is None else root_password

    root_client = (
        client
        if is_admin(client)
        else client.login(email=root_email, password=root_password)
    )
    if name is None and path is None:
        raise SyftException("Please provide either a checkpoint name or a path.")

    checkpoint_zip_path = get_checkpoint_for(path=path, chkpt_name=name)

    if checkpoint_zip_path is None:
        print(f"No last checkpoint found for : {name} or {path}")
        return

    # Unzip the checkpoint zip file
    with zipfile.ZipFile(checkpoint_zip_path, "r") as zipf:
        checkpoint_temp_dir = Path(tempfile.mkdtemp())
        zipf.extract("checkpoint.blob", checkpoint_temp_dir)
        zipf.extract("checkpoint.yaml", checkpoint_temp_dir)

    checkpoint_blob = checkpoint_temp_dir / "checkpoint.blob"

    print(f"Loading from checkpoint: {checkpoint_zip_path}")
    result = root_client.load_migration_data(
        path_or_data=checkpoint_blob,
        include_worker_pools=True,
        with_reset_db=True,
    )

    if isinstance(result, SyftError):
        raise SyftException(message=result.message)

    print("Successfully loaded data from checkpoint.")

    # Step 1: Build and push the worker images

    print("Recreating worker images from checkpoint.")
    worker_image_list = (
        [] if root_client.images.get_all() is None else root_client.images.get_all()
    )
    for worker_image in worker_image_list:
        if worker_image.is_prebuilt:
            continue

        registry = worker_image.image_identifier.registry

        build_and_push_image(
            root_client,
            worker_image,
            registry_uid=registry.id if registry else None,
            tag=worker_image.image_identifier.repo_with_tag,
            reg_password=registry_username,
            reg_username=registry_password,
            force_build=True,
        )

    print("Successfully Built worker image data from checkpoint.")

    # Step 2: Recreate the worker pools
    print("Recreating worker pools from checkpoint.")
    worker_pool_list = (
        [] if root_client.worker_pools is None else root_client.worker_pools
    )
    for worker_pool in worker_pool_list:
        previous_worker_cnt = worker_pool.max_count
        purge_res = root_client.worker_pools.purge_workers(pool_id=worker_pool.id)
        print(purge_res)
        add_res = root_client.worker_pools.add_workers(
            number=previous_worker_cnt,
            pool_id=worker_pool.id,
            registry_username=registry_username,
            registry_password=registry_password,
        )
        print(add_res)

    print("Successfully loaded worker pool data from checkpoint.")



# relative
from ...serde.serializable import serializable
from ...server.credentials import SyftVerifyKey
from ...store.db.stash import ObjectStash
from ...store.document_store_errors import StashException
from ...types.result import as_result
from .code_history import CodeHistory


@serializable(canonical_name="CodeHistoryStashSQL", version=1)
class CodeHistoryStash(ObjectStash[CodeHistory]):
    @as_result(StashException)
    def get_by_service_func_name_and_verify_key(
        self,
        credentials: SyftVerifyKey,
        service_func_name: str,
        user_verify_key: SyftVerifyKey,
    ) -> CodeHistory:
        return self.get_one(
            credentials=credentials,
            filters={
                "user_verify_key": user_verify_key,
                "service_func_name": service_func_name,
            },
        ).unwrap()

    @as_result(StashException)
    def get_by_service_func_name(
        self, credentials: SyftVerifyKey, service_func_name: str
    ) -> list[CodeHistory]:
        return self.get_all(
            credentials=credentials,
            filters={"service_func_name": service_func_name},
        ).unwrap()

    @as_result(StashException)
    def get_by_verify_key(
        self, credentials: SyftVerifyKey, user_verify_key: SyftVerifyKey
    ) -> list[CodeHistory]:
        return self.get_all(
            credentials=credentials,
            filters={"user_verify_key": user_verify_key},
        ).unwrap()



# Can also be specified by the environment variable
# ORCHESTRA_DEPLOYMENT_TYPE
# stdlib
from enum import Enum

# relative
from .serde.serializable import serializable
from .types.syft_object import SYFT_OBJECT_VERSION_1


@serializable()
class DeploymentType(str, Enum):
    __canonical_name__ = "DeploymentType"
    __version__ = SYFT_OBJECT_VERSION_1

    PYTHON = "python"
    REMOTE = "remote"

    def __str__(self) -> str:
        # Use values when transforming ServerType to str
        return self.value



# stdlib
from datetime import datetime
from datetime import timedelta
import secrets
import string
from typing import TypeVar
from typing import cast

# relative
from ...abstract_server import ServerType
from ...serde.serializable import serializable
from ...server.credentials import SyftSigningKey
from ...server.credentials import SyftVerifyKey
from ...store.db.db import DBManager
from ...store.document_store_errors import NotFoundException
from ...store.document_store_errors import StashException
from ...store.linked_obj import LinkedObject
from ...types.errors import CredentialsError
from ...types.errors import SyftException
from ...types.result import as_result
from ...types.syft_metaclass import Empty
from ...types.uid import UID
from ..action.action_permissions import ActionObjectPermission
from ..action.action_permissions import ActionPermission
from ..context import AuthedServiceContext
from ..context import ServerServiceContext
from ..context import UnauthedServiceContext
from ..notification.email_templates import OnBoardEmailTemplate
from ..notification.email_templates import PasswordResetTemplate
from ..notification.notification_service import CreateNotification
from ..notifier.notifier_enums import NOTIFIERS
from ..response import SyftSuccess
from ..service import AbstractService
from ..service import SERVICE_TO_TYPES
from ..service import TYPE_TO_SERVICE
from ..service import service_method
from ..settings.settings import PwdTokenResetConfig
from ..settings.settings_stash import SettingsStash
from .errors import UserEnclaveAdminLoginError
from .errors import UserError
from .errors import UserPermissionError
from .errors import UserUpdateError
from .user import User
from .user import UserCreate
from .user import UserPrivateKey
from .user import UserSearch
from .user import UserUpdate
from .user import UserView
from .user import check_pwd
from .user import salt_and_hash_password
from .user import validate_password
from .user_roles import ADMIN_ROLE_LEVEL
from .user_roles import DATA_OWNER_ROLE_LEVEL
from .user_roles import DATA_SCIENTIST_ROLE_LEVEL
from .user_roles import GUEST_ROLE_LEVEL
from .user_roles import ServiceRole
from .user_roles import ServiceRoleCapability
from .user_stash import UserStash

T = TypeVar("T")


def _paginate(
    list_objs: list[T], page_size: int | None = 0, page_index: int | None = 0
) -> list[T]:
    # If chunk size is defined, then split list into evenly sized chunks
    if page_size:
        _list_objs = [
            list_objs[i : i + page_size] for i in range(0, len(list_objs), page_size)
        ]

        # Return the proper slice using chunk_index
        if page_index is not None:
            _list_objs = _list_objs[page_index]  # type: ignore
        else:
            _list_objs = _list_objs[0]  # type: ignore
        return _list_objs  # type: ignore

    return list_objs


@serializable(canonical_name="UserService", version=1)
class UserService(AbstractService):
    stash: UserStash

    def __init__(self, store: DBManager) -> None:
        self.stash = UserStash(store=store)

    @as_result(StashException)
    def _add_user(self, credentials: SyftVerifyKey, user: User) -> User:
        return self.stash.set(
            credentials=credentials,
            obj=user,
            add_permissions=[
                ActionObjectPermission(
                    uid=user.id, permission=ActionPermission.ALL_READ
                ),
            ],
        ).unwrap()

    def _check_if_email_exists(self, credentials: SyftVerifyKey, email: str) -> bool:
        try:
            self.stash.get_by_email(credentials=credentials, email=email).unwrap()
            return True
        except NotFoundException:
            return False

    @service_method(path="user.create", name="create", autosplat="user_create")
    def create(
        self, context: AuthedServiceContext, user_create: UserCreate
    ) -> UserView:
        """Create a new user"""
        user = user_create.to(User)

        user_exists = self._check_if_email_exists(
            credentials=context.credentials, email=user.email
        )

        # TODO: Ensure we don't leak out the existence of a user
        if user_exists:
            raise SyftException(public_message=f"User {user.email} already exists")

        new_user = self._add_user(context.credentials, user).unwrap()
        return new_user.to(UserView)

    def forgot_password(
        self, context: UnauthedServiceContext, email: str
    ) -> SyftSuccess:
        success_msg = (
            "If the email is valid, we sent a password "
            + "reset token to your email or a password request to the admin."
        )
        root_key = self.root_verify_key

        root_context = AuthedServiceContext(server=context.server, credentials=root_key)

        result = self.stash.get_by_email(credentials=root_key, email=email)

        # Isn't a valid email
        if result.is_err():
            return SyftSuccess(message=success_msg)
        user = result.ok()

        if user is None:
            return SyftSuccess(message=success_msg)

        user_role = self.get_role_for_credentials(user.verify_key).unwrap()
        if user_role == ServiceRole.ADMIN:
            raise SyftException(
                public_message="You can't request password reset for an Admin user."
            )

        # Email is valid
        # Notifications disabled
        # We should just sent a notification to the admin/user about password reset
        # Notifications Enabled
        # Instead of changing the password here, we would change it in email template generation.
        link = LinkedObject.with_context(user, context=root_context)
        # Notifier is active
        notifier = root_context.server.services.notifier.settings(
            context=root_context
        ).unwrap()
        notification_is_enabled = notifier.active
        # Email is enabled
        email_is_enabled = notifier.email_enabled
        # User Preferences allow email notification
        user_allow_email_notifications = user.notifications_enabled[NOTIFIERS.EMAIL]

        # This checks if the user will safely receive the email reset.
        not_receive_emails = (
            not notification_is_enabled
            or not email_is_enabled
            or not user_allow_email_notifications
        )

        # If notifier service is not enabled.
        if not_receive_emails:
            message = CreateNotification(
                subject="You requested password reset.",
                from_user_verify_key=root_key,
                to_user_verify_key=user.verify_key,
                linked_obj=link,
            )
            result = root_context.server.services.notification.send(
                context=root_context, notification=message
            )
            message = CreateNotification(
                subject="User requested password reset.",
                from_user_verify_key=user.verify_key,
                to_user_verify_key=root_key,
                linked_obj=link,
            )

            result = root_context.server.services.notification.send(
                context=root_context, notification=message
            )
        else:
            # Email notification is Enabled
            # Therefore, we can directly send a message to the
            # user with its new password.
            message = CreateNotification(
                subject="You requested a password reset.",
                from_user_verify_key=root_key,
                to_user_verify_key=user.verify_key,
                linked_obj=link,
                notifier_types=[NOTIFIERS.EMAIL],
                email_template=PasswordResetTemplate,
            )
            result = root_context.server.services.notification.send(
                context=root_context, notification=message
            )

        return SyftSuccess(message=success_msg)

    @service_method(
        path="user.request_password_reset",
        name="request_password_reset",
        roles=ADMIN_ROLE_LEVEL,
    )
    def request_password_reset(self, context: AuthedServiceContext, uid: UID) -> str:
        user = self.stash.get_by_uid(credentials=context.credentials, uid=uid).unwrap()
        user_role = self.get_role_for_credentials(user.verify_key).unwrap()

        if user_role == ServiceRole.ADMIN:
            raise SyftException(
                public_message="You can't request password reset for an Admin user."
            )

        user.reset_token = self.generate_new_password_reset_token(
            context.server.settings.pwd_token_config
        )
        user.reset_token_date = datetime.now()

        self.stash.update(
            credentials=context.credentials, obj=user, has_permission=True
        ).unwrap()

        return user.reset_token

    def reset_password(
        self, context: UnauthedServiceContext, token: str, new_password: str
    ) -> SyftSuccess:
        """Resets a certain user password using a temporary token."""
        root_key = self.root_verify_key

        root_context = AuthedServiceContext(server=context.server, credentials=root_key)
        try:
            user = self.stash.get_by_reset_token(
                credentials=root_context.credentials, token=token
            ).unwrap()
        except NotFoundException:
            raise SyftException(
                public_message="Failed to reset user password. Token is invalid or expired."
            )
        #
        if user is None:
            raise SyftException(
                public_message="Failed to reset user password. Token is invalid or expired."
            )
        now = datetime.now()
        if user.reset_token_date is not None:
            time_difference = now - user.reset_token_date
        else:
            raise SyftException(
                public_message="Failed to reset user password. Reset Token Invalid!"
            )

        # If token expired
        expiration_time = root_context.server.settings.pwd_token_config.token_exp_min
        if time_difference > timedelta(seconds=expiration_time):
            raise SyftException(
                public_message="Failed to reset user password. Token is invalid or expired."
            )

        if not validate_password(new_password):
            raise SyftException(
                public_message=(
                    "Your new password must have at least 8 characters, an upper case "
                    "and lower case character; and at least one number."
                )
            )

        salt, hashed = salt_and_hash_password(new_password, 12)
        user.hashed_password = hashed
        user.salt = salt

        user.reset_token = None
        user.reset_token_date = None

        self.stash.update(
            credentials=root_context.credentials, obj=user, has_permission=True
        ).unwrap()

        return SyftSuccess(message="User Password updated successfully.")

    def generate_new_password_reset_token(
        self, token_config: PwdTokenResetConfig
    ) -> str:
        valid_characters = ""
        if token_config.ascii:
            valid_characters += string.ascii_letters

        if token_config.numbers:
            valid_characters += string.digits

        generated_token = "".join(
            secrets.choice(valid_characters) for _ in range(token_config.token_len)
        )

        return generated_token

    @service_method(path="user.view", name="view", roles=DATA_SCIENTIST_ROLE_LEVEL)
    def view(self, context: AuthedServiceContext, uid: UID) -> UserView:
        """Get user for given uid"""
        user = self.stash.get_by_uid(credentials=context.credentials, uid=uid).unwrap()
        return user.to(UserView)

    @service_method(path="user.get_all", name="get_all", roles=DATA_OWNER_ROLE_LEVEL)
    def get_all(
        self,
        context: AuthedServiceContext,
        order_by: str | None = None,
        sort_order: str | None = None,
        page_size: int | None = 0,
        page_index: int | None = 0,
    ) -> list[UserView]:
        users = self.stash.get_all(
            context.credentials,
            order_by=order_by,
            sort_order=sort_order,
        ).unwrap()
        users = [user.to(UserView) for user in users]
        return _paginate(users, page_size, page_index)

    @service_method(
        path="user.get_index", name="get_index", roles=DATA_OWNER_ROLE_LEVEL
    )
    def get_index(
        self,
        context: AuthedServiceContext,
        index: int,
    ) -> UserView:
        return (
            self.stash.get_index(credentials=context.credentials, index=index)
            .unwrap()
            .to(UserView)
        )

    def signing_key_for_verify_key(self, verify_key: SyftVerifyKey) -> UserPrivateKey:
        user = self.stash.get_by_verify_key(
            credentials=self.stash.root_verify_key, verify_key=verify_key
        ).unwrap()

        return user.to(UserPrivateKey)

    @as_result(SyftException)
    def get_role_for_credentials(
        self, credentials: SyftVerifyKey | SyftSigningKey
    ) -> ServiceRole:
        try:
            # they could be different
            # TODO: This fn is cryptic -- when does each situation occur?
            if isinstance(credentials, SyftVerifyKey):
                role = self.stash.get_role(credentials=credentials)
                return role
            elif isinstance(credentials, SyftSigningKey):
                user = self.stash.get_by_signing_key(
                    credentials=credentials.verify_key,
                    signing_key=credentials,  # type: ignore
                ).unwrap()
            else:
                raise CredentialsError
        except NotFoundException:
            return ServiceRole.GUEST

        return cast(ServiceRole, user.role)

    @service_method(path="user.search", name="search", autosplat=["user_search"])
    def search(
        self,
        context: AuthedServiceContext,
        user_search: UserSearch,
        page_size: int | None = 0,
        page_index: int | None = 0,
    ) -> list[UserView]:
        kwargs = user_search.to_dict(exclude_empty=True)
        kwargs.pop("created_date")
        kwargs.pop("updated_date")
        kwargs.pop("deleted_date")
        if len(kwargs) == 0:
            raise SyftException(public_message="Invalid search parameters")

        users = self.stash.get_all(
            credentials=context.credentials, filters=kwargs
        ).unwrap()

        users = [user.to(UserView) for user in users] if users is not None else []
        return _paginate(users, page_size, page_index)

    @as_result(StashException, NotFoundException)
    def get_user_id_for_credentials(self, credentials: SyftVerifyKey) -> UID:
        user = self.stash.get_by_verify_key(
            credentials=credentials, verify_key=credentials
        ).unwrap()
        return cast(UID, user.id)

    @service_method(
        path="user.get_current_user", name="get_current_user", roles=GUEST_ROLE_LEVEL
    )
    def get_current_user(self, context: AuthedServiceContext) -> UserView:
        user = self.stash.get_by_verify_key(
            credentials=context.credentials, verify_key=context.credentials
        ).unwrap()
        return user.to(UserView)

    @service_method(
        path="user.get_by_verify_key", name="get_by_verify_key", roles=ADMIN_ROLE_LEVEL
    )
    def get_by_verify_key_endpoint(
        self, context: AuthedServiceContext, verify_key: SyftVerifyKey
    ) -> UserView:
        user = self.stash.get_by_verify_key(
            credentials=context.credentials, verify_key=verify_key
        ).unwrap()
        return user.to(UserView)

    @service_method(
        path="user.update",
        name="update",
        roles=GUEST_ROLE_LEVEL,
        autosplat="user_update",
    )
    def update(
        self, context: AuthedServiceContext, uid: UID, user_update: UserUpdate
    ) -> UserView:
        updates_role = user_update.role is not Empty  # type: ignore[comparison-overlap]
        can_edit_roles = ServiceRoleCapability.CAN_EDIT_ROLES in context.capabilities()

        if updates_role and not can_edit_roles:
            raise UserPermissionError(
                f"User {context.credentials} tried to update user {uid} with {user_update}."
            )

        if (user_update.mock_execution_permission is not Empty) and not can_edit_roles:  # type: ignore[comparison-overlap]
            raise UserPermissionError(
                f"User {context.credentials} with role {context.role} is not allowed"
                " to update permissions."
            )

        # Get user to be updated by its UID
        user = self.stash.get_by_uid(credentials=context.credentials, uid=uid).unwrap()

        immutable_fields = {"created_date", "updated_date", "deleted_date"}
        updated_fields = user_update.to_dict(
            exclude_none=True, exclude_empty=True
        ).keys()

        for field_name in immutable_fields:
            if field_name in updated_fields:
                raise SyftException(
                    public_message=f"You are not allowed to modify '{field_name}'."
                )

        # important to prevent root admins from shooting themselves in the foot
        if (
            user_update.role is not Empty  # type: ignore
            and user.verify_key == context.server.verify_key
        ):
            raise SyftException(public_message="Cannot update root role")

        if (
            user_update.verify_key is not Empty
            and user.verify_key == context.server.verify_key
        ):
            raise SyftException(public_message="Cannot update root verify key")

        if user_update.name is not Empty and user_update.name.strip() == "":  # type: ignore[comparison-overlap]
            raise SyftException(public_message="Name can't be an empty string.")

        # check if the email already exists (with root's key)
        if user_update.email is not Empty:
            user_exists = self.stash.email_exists(email=user_update.email).unwrap()
            if user_exists:
                raise UserUpdateError(
                    public_message=f"User {user_update.email} already exists"
                )

        if updates_role:
            if context.role == ServiceRole.ADMIN:
                # do anything
                pass
            elif (
                context.role == ServiceRole.DATA_OWNER
                and user.role is not None
                and context.role.value > user.role.value
                and context.role.value > user_update.role.value
            ):
                # as a data owner, only update lower roles to < data owner
                pass
            else:
                raise UserPermissionError(
                    f"User {context.credentials} tried to update user {uid}"
                    f" with {user_update}."
                )

        edits_non_role_attrs = any(
            getattr(user_update, attr) is not Empty
            for attr in user_update.to_dict()
            if attr not in ["role", "created_date", "updated_date", "deleted_date"]
        )
        if (
            edits_non_role_attrs
            and user.verify_key != context.credentials
            and ServiceRoleCapability.CAN_MANAGE_USERS not in context.capabilities()
        ):
            raise UserPermissionError(
                f"User {context.credentials} tried to update user {uid}"
                f" with {user_update}."
            )

        # Fill User Update fields that will not be changed by replacing it
        # for the current values found in user obj.
        for name, value in user_update.to_dict(exclude_empty=True).items():
            if name == "password" and value:
                salt, hashed = salt_and_hash_password(value, 12)
                user.hashed_password = hashed
                user.salt = salt
            elif not name.startswith("__") and value is not None:
                setattr(user, name, value)

        user = self.stash.update(
            credentials=context.credentials, obj=user, has_permission=True
        ).unwrap()

        if user.role == ServiceRole.ADMIN:
            settings_stash = SettingsStash(store=self.stash.db)
            settings = settings_stash.get_all(
                context.credentials, limit=1, sort_order="desc"
            ).unwrap()

            # TODO: Chance to refactor here in settings, as we're always doing get_att[0]
            if len(settings) > 0:
                settings_data = settings[0]
                settings_data.admin_email = user.email
                settings_stash.update(
                    credentials=context.credentials, obj=settings_data
                )

        return user.to(UserView)

    @service_method(path="user.delete", name="delete", roles=GUEST_ROLE_LEVEL)
    def delete(self, context: AuthedServiceContext, uid: UID) -> UID:
        user_to_delete = self.stash.get_by_uid(
            credentials=context.credentials, uid=uid
        ).unwrap()

        # Cannot delete root user
        if user_to_delete.verify_key == self.root_verify_key:
            raise UserPermissionError(
                private_message=f"User {context.credentials} attempted to delete root user."
            )

        # - Admins can delete any user
        # - Data Owners can delete Data Scientists and Guests
        has_delete_permissions = (
            context.role == ServiceRole.ADMIN
            or context.role == ServiceRole.DATA_OWNER
            and user_to_delete.role in [ServiceRole.GUEST, ServiceRole.DATA_SCIENTIST]
        )

        if not has_delete_permissions:
            raise UserPermissionError(
                private_message=(
                    f"User {context.credentials} ({context.role}) tried to delete user "
                    f"{uid} ({user_to_delete.role})"
                )
            )

        # TODO: Remove notifications for the deleted user
        return self.stash.delete_by_uid(
            credentials=context.credentials, uid=uid
        ).unwrap()

    def exchange_credentials(self, context: UnauthedServiceContext) -> SyftSuccess:
        """Verify user
        TODO: We might want to use a SyftObject instead
        """

        if context.login_credentials is None:
            raise SyftException(public_message="Invalid login credentials")

        user = self.stash.get_by_email(
            credentials=self.root_verify_key, email=context.login_credentials.email
        ).unwrap()

        if check_pwd(context.login_credentials.password, user.hashed_password):
            if (
                context.server
                and context.server.server_type == ServerType.ENCLAVE
                and user.role == ServiceRole.ADMIN
            ):
                # FIX: Replace with SyftException
                raise SyftException(
                    public_message=UserEnclaveAdminLoginError.public_message
                )
        else:
            # FIX: Replace this below
            raise SyftException(public_message=CredentialsError.public_message)

        return SyftSuccess(message="Login successful.", value=user.to(UserPrivateKey))

    @property
    def root_verify_key(self) -> SyftVerifyKey:
        return self.stash.root_verify_key

    def register(
        self, context: ServerServiceContext, new_user: UserCreate
    ) -> SyftSuccess:
        """Register new user"""

        # this method handles errors in a slightly different way as it is directly called instead of
        # going through Server.handle_message

        request_user_role = (
            ServiceRole.GUEST
            if new_user.created_by is None
            else self.get_role_for_credentials(new_user.created_by).unwrap()
        )

        can_user_register = (
            context.server.settings.signup_enabled
            or request_user_role in DATA_OWNER_ROLE_LEVEL
        )

        if not can_user_register:
            raise SyftException(
                public_message="You have no permission to create an account. Please contact the Datasite owner."
            )

        user = new_user.to(User)

        user_exists = self._check_if_email_exists(
            credentials=user.verify_key, email=user.email
        )

        if user_exists:
            raise SyftException(public_message=f"User {user.email} already exists")

        user = self._add_user(credentials=user.verify_key, user=user).unwrap(
            public_message=f"Failed to create user {user.email}"
        )
        success_message = f"User '{user.name}' successfully registered!"

        # Notification Step
        root_key = self.root_verify_key
        root_context = AuthedServiceContext(server=context.server, credentials=root_key)
        link = None

        if new_user.created_by:
            link = LinkedObject.with_context(user, context=root_context)

        message = CreateNotification(
            subject=success_message,
            from_user_verify_key=root_key,
            to_user_verify_key=user.verify_key,
            linked_obj=link,
            notifier_types=[NOTIFIERS.EMAIL],
            email_template=OnBoardEmailTemplate,
        )
        context.server.services.notification.send(
            context=root_context, notification=message
        )

        if request_user_role in DATA_OWNER_ROLE_LEVEL:
            success_message += " To see users, run `[your_client].users`"

        return SyftSuccess(message=success_message, value=user.to(UserPrivateKey))

    @as_result(StashException)
    def user_verify_key(self, email: str) -> SyftVerifyKey:
        # we are bypassing permissions here, so dont use to return a result directly to the user
        credentials = self.root_verify_key
        user = self.stash.get_by_email(credentials=credentials, email=email).unwrap()
        if user.verify_key is None:
            raise UserError(f"User {email} has no verify key")
        return user.verify_key

    @as_result(StashException)
    def get_by_verify_key(self, verify_key: SyftVerifyKey) -> UserView:
        # we are bypassing permissions here, so dont use to return a result directly to the user
        credentials = self.root_verify_key
        user = self.stash.get_by_verify_key(
            credentials=credentials, verify_key=verify_key
        ).unwrap()
        return user.to(UserView)

    @as_result(StashException)
    def _set_notification_status(
        self,
        notifier_type: NOTIFIERS,
        new_status: bool,
        verify_key: SyftVerifyKey,
    ) -> None:
        user = self.stash.get_by_verify_key(
            credentials=verify_key, verify_key=verify_key
        ).unwrap()
        user.notifications_enabled[notifier_type] = new_status
        self.stash.update(credentials=user.verify_key, obj=user).unwrap()

    @as_result(SyftException)
    def enable_notifications(
        self, context: AuthedServiceContext, notifier_type: NOTIFIERS
    ) -> SyftSuccess:
        self._set_notification_status(
            notifier_type=notifier_type, new_status=True, verify_key=context.credentials
        ).unwrap()
        return SyftSuccess(message="Notifications enabled successfully!")

    def disable_notifications(
        self, context: AuthedServiceContext, notifier_type: NOTIFIERS
    ) -> SyftSuccess:
        self._set_notification_status(
            notifier_type=notifier_type,
            new_status=False,
            verify_key=context.credentials,
        ).unwrap()

        return SyftSuccess(message="Notifications disabled successfully!")


TYPE_TO_SERVICE[User] = UserService
SERVICE_TO_TYPES[UserService].update({User})



# third party
from pydantic import Field

# relative
from ..serde.serializable import serializable
from .document_store import DocumentStore
from .document_store import StoreClientConfig
from .document_store import StoreConfig
from .kv_document_store import KeyValueBackingStore
from .kv_document_store import KeyValueStorePartition
from .locks import LockingConfig
from .locks import NoLockingConfig


@serializable(
    attrs=["index_name", "settings", "store_config"],
    canonical_name="SQLiteBackingStore",
    version=1,
)
class SQLiteBackingStore(KeyValueBackingStore):
    """Core Store logic for the SQLite stores."""

    pass


@serializable(canonical_name="SQLiteStorePartition", version=1)
class SQLiteStorePartition(KeyValueStorePartition):
    """SQLite StorePartition
    Parameters:
        `settings`: PartitionSettings
            PySyft specific settings, used for indexing and partitioning
        `store_config`: SQLiteStoreConfig
            SQLite specific configuration
    """


# the base document store is already a dict but we can change it later
@serializable(canonical_name="SQLiteDocumentStore", version=1)
class SQLiteDocumentStore(DocumentStore):
    """SQLite Document Store
    Parameters:
        `store_config`: StoreConfig
            SQLite specific configuration, including connection details and client class type.
    """

    partition_type = SQLiteStorePartition


@serializable(canonical_name="SQLiteStoreClientConfig", version=1)
class SQLiteStoreClientConfig(StoreClientConfig):
    """SQLite connection config"""

    pass


@serializable()
class SQLiteStoreConfig(StoreConfig):
    __canonical_name__ = "SQLiteStoreConfig"
    """SQLite Store config, used by SQLiteStorePartition
    Parameters:
        `client_config`: SQLiteStoreClientConfig
            SQLite connection configuration
        `store_type`: DocumentStore
            Class interacting with QueueStash. Default: SQLiteDocumentStore
        `backing_store`: KeyValueBackingStore
            The Store core logic. Default: SQLiteBackingStore
        locking_config: LockingConfig
            The config used for store locking. Available options:
                * NoLockingConfig: no locking, ideal for single-thread stores.
                * ThreadingLockingConfig: threading-based locking, ideal for same-process in-memory stores.
            Defaults to NoLockingConfig.
    """

    client_config: SQLiteStoreClientConfig
    store_type: type[DocumentStore] = SQLiteDocumentStore
    backing_store: type[KeyValueBackingStore] = SQLiteBackingStore
    locking_config: LockingConfig = Field(default_factory=NoLockingConfig)



# relative
from ..serde.serializable import serializable
from .document_store import StoreClientConfig


@serializable(canonical_name="MongoStoreClientConfig", version=1)
class MongoStoreClientConfig(StoreClientConfig):
    pass



# stdlib
from copy import deepcopy
import traceback
from typing import Any
from typing import TYPE_CHECKING

# third party
from typing_extensions import Self

# relative
from ..serde.serializable import serializable
from ..types.base import SyftBaseModel
from ..util.util import sanitize_html

if TYPE_CHECKING:
    # relative
    from .context import AuthedServiceContext


class SyftResponseMessage(SyftBaseModel):
    message: str
    _bool: bool = True
    require_api_update: bool = False
    client_warnings: list[str] = []

    def add_warnings_from_context(self, context: "AuthedServiceContext") -> None:
        self.client_warnings = deepcopy(context.client_warnings)

    def is_err(self) -> bool:
        return False

    def is_ok(self) -> bool:
        return True

    def __getattr__(self, name: str) -> Any:
        if name in [
            "_bool",
            # "_repr_html_",
            # "message",
            # 'require_api_update',
            # '__bool__',
            # '__eq__',
            # '__repr__',
            # '__str__',
            # '_repr_html_class_',
            # '_repr_html_',
            "_ipython_canary_method_should_not_exist_",
            "_ipython_display_",
            "__canonical_name__",
            "__version__",
        ] or name.startswith("_repr"):
            return super().__getattr__(name)
        raise AttributeError(
            f"You have tried accessing `{name}` on a {type(self).__name__} with message: {self.message}"
        )

    def __bool__(self) -> bool:
        return self._bool

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, SyftResponseMessage):
            return (
                self.message == other.message
                and self._bool == other._bool
                and self.require_api_update == other.require_api_update
            )
        return self._bool == other

    def __repr__(self) -> str:
        _class_name_ = type(self).__name__
        return f"{_class_name_}: {self.message}"

    def __str__(self) -> str:
        return self.__repr__()

    @property
    def _repr_html_class_(self) -> str:
        return "alert-info"

    def _repr_html_(self) -> str:
        return (
            f'<div class="{self._repr_html_class_}">'
            f"<strong>{type(self).__name__}</strong>: "
            f'<pre class="{self._repr_html_class_}">'
            f"{sanitize_html(self.message)}</pre></div><br/>"
        )


@serializable(canonical_name="SyftError", version=1)
class SyftError(SyftResponseMessage):
    _bool: bool = False
    tb: str | None = None

    @property
    def _repr_html_class_(self) -> str:
        return "alert-danger"

    def __bool__(self) -> bool:
        return False

    def is_err(self) -> bool:
        return True

    def is_ok(self) -> bool:
        return False

    @classmethod
    def from_public_exception(
        cls,
        exc: Exception,
    ) -> Self:
        return cls(message=exc.public_message)

    @classmethod
    def from_exception(
        cls,
        context: "AuthedServiceContext",
        exc: Exception,
        include_traceback: bool = False,
    ) -> Self:
        # traceback may contain private information
        # relative
        from ..types.errors import SyftException as NewSyftException

        tb = None

        if include_traceback:
            if isinstance(exc, NewSyftException):
                error_msg = exc.get_message(context)
                tb = exc.get_tb(context)
            else:
                # other exceptions
                lines = traceback.format_exception(exc)
                tb = "".join(lines)
                error_msg = lines[-1]
                print(f"Error: {tb}")
        else:
            if isinstance(exc, NewSyftException):
                error_msg = exc.get_message(context)
            else:
                # by default only type
                error_msg = f"Something unexpected happened server side {type(exc)}"
            print(f"Error: {exc}")
            print(traceback.format_exc())
        return cls(message=error_msg, tb=tb)


@serializable(canonical_name="SyftSuccess", version=1)
class SyftSuccess(SyftResponseMessage):
    value: Any | None = None

    def is_err(self) -> bool:
        return False

    def is_ok(self) -> bool:
        return True

    @property
    def _repr_html_class_(self) -> str:
        return "alert-success"

    def unwrap_value(self) -> Any:
        return self.value


@serializable(canonical_name="SyftNotReady", version=1)
class SyftNotReady(SyftError):
    _bool: bool = False

    @property
    def _repr_html_class_(self) -> str:
        return "alert-info"


@serializable(canonical_name="SyftWarning", version=1)
class SyftWarning(SyftResponseMessage):
    @property
    def _repr_html_class_(self) -> str:
        return "alert-warning"


@serializable(canonical_name="SyftInfo", version=1)
class SyftInfo(SyftResponseMessage):
    _bool: bool = False

    @property
    def _repr_html_class_(self) -> str:
        return "alert-info"



# stdlib
import math
from typing import Any


def _safe_isnan(v: Any) -> bool:
    try:
        return math.isnan(v)
    except TypeError:
        return False


def _syft_equals(v1: Any, v2: Any) -> bool:
    if v1 is None:
        return v2 is None

    # handle nan, since nan==nan is False
    if _safe_isnan(v1):
        return _safe_isnan(v2)

    if isinstance(v1, dict) and isinstance(v2, dict):
        return _syft_recursive_dict_equals(v1, v2)

    return v1 == v2


def _syft_recursive_dict_equals(d1: dict, d2: dict) -> bool:
    if len(d1) != len(d2):
        return False

    for k in d1.keys():
        if k not in d2:
            return False
        if not _syft_equals(d1[k], d2[k]):
            return False

    return True



# relative
from ...serde.serializable import serializable
from ...store.db.stash import ObjectStash
from ...types.blob_storage import BlobStorageEntry


@serializable(canonical_name="BlobStorageSQLStash", version=1)
class BlobStorageStash(ObjectStash[BlobStorageEntry]):
    pass



# stdlib
from dataclasses import dataclass
from dataclasses import field
import json
import re
import time
from typing import Any

# third party
from aiosmtpd.controller import Controller
from faker import Faker
from filelock import FileLock

# relative
from ...service.user.user_roles import ServiceRole

fake = Faker()


@dataclass
class Email:
    email_from: str
    email_to: str
    email_content: str

    def to_dict(self) -> dict:
        output = {}
        for k, v in self.__dict__.items():
            output[k] = v
        return output

    def __iter__(self):
        yield from self.to_dict().items()

    def __getitem__(self, key):
        return self.to_dict()[key]

    def __repr__(self) -> str:
        return f"{self.email_to}\n{self.email_from}\n\n{self.email_content}"


class EmailServer:
    def __init__(self, filepath="emails.json"):
        self.filepath = filepath
        lockpath = self.filepath + ".lock"
        self._lock = FileLock(lock_file=lockpath)
        self._emails: dict[str, list[Email]] = self.load_emails()

    def load_emails(self) -> dict[str, list[Email]]:
        try:
            with (
                self._lock as _,
                open(self.filepath) as f,
            ):
                data = json.load(f)
                return {k: [Email(**email) for email in v] for k, v in data.items()}
        except Exception as e:
            print("Issues reading email file. Using empty email dict.", e)
            return {}

    def save_emails(self) -> None:
        with (
            self._lock as _,
            open(self.filepath, "w") as f,
        ):
            data = {
                k: [email.to_dict() for email in v] for k, v in self._emails.items()
            }
            f.write(json.dumps(data))

    def add_email_for_user(self, user_email: str, email: Email) -> None:
        with self._lock:
            if user_email not in self._emails:
                self._emails[user_email] = []
            self._emails[user_email].append(email)
            self.save_emails()

    def get_emails_for_user(self, user_email: str) -> list[Email]:
        self._emails: dict[str, list[Email]] = self.load_emails()
        return self._emails.get(user_email, [])

    def reset_emails(self) -> None:
        with self._lock:
            self._emails = {}
            self.save_emails()


SENDER = "noreply@openmined.org"


def get_token(email) -> str:
    # stdlib
    import re

    pattern = r"syft_client\.reset_password\(token='(.*?)', new_password=.*?\)"
    try:
        token = re.search(pattern, email.email_content).group(1)
    except Exception:
        raise Exception(f"No token found in email: {email.email_content}")
    return token


@dataclass
class TestUser:
    name: str
    email: str
    password: str
    role: ServiceRole
    new_password: str | None = None
    email_disabled: bool = False
    reset_password: bool = False
    reset_token: str | None = None
    _client_cache: Any | None = field(default=None, repr=False, init=False)
    _email_server: EmailServer | None = None

    @property
    def latest_password(self) -> str:
        if self.new_password:
            return self.new_password
        return self.password

    def make_new_password(self) -> str:
        self.new_password = fake.password()
        return self.new_password

    @property
    def client(self):
        return self._client_cache

    def relogin(self) -> None:
        self.client = self.client

    @client.setter
    def client(self, client):
        this_client = client.login(email=self.email, password=self.latest_password)
        self._client_cache = this_client

    def to_dict(self) -> dict:
        output = {}
        for k, v in self.__dict__.items():
            if k.startswith("_"):
                continue
            if k == "role":
                v = str(v)
            output[k] = v
        return output

    def __iter__(self):
        for key, val in self.to_dict().items():
            if not key.startswith("_"):
                yield key, val

    def __getitem__(self, key):
        if key.startswith("_"):
            return None
        return self.to_dict()[key]

    def update_password(self):
        self.password = self.new_password
        self.new_password = None

    @property
    def emails(self) -> list[Email]:
        if not self._email_server:
            print("Not connected to email server object")
            return []
        return self._email_server.get_emails_for_user(self.email)

    def get_token(self) -> str:
        for email in reversed(self.emails):
            token = None
            try:
                token = get_token(email)
                break
            except Exception:  # nosec
                pass
        self.reset_token = token
        return token


def save_users(users):
    user_dicts = []
    for user in users:
        user_dicts.append(user.to_dict())
    print(user_dicts)
    with open("./users.json", "w") as f:
        f.write(json.dumps(user_dicts))


def load_users(high_client: None, path="./users.json"):
    users = []
    with open(path) as f:
        data = f.read()
        user_dicts = json.loads(data)
    for user in user_dicts:
        test_user = TestUser(**user)
        if high_client:
            test_user.client = high_client
        users.append(test_user)
    return users


def make_user(
    name: str | None = None,
    email: str | None = None,
    password: str | None = None,
    role: ServiceRole = ServiceRole.DATA_SCIENTIST,
):
    fake = Faker()
    if name is None:
        name = fake.name()
    if email is None:
        ascii_string = re.sub(r"[^a-zA-Z\s]", "", name).lower()
        dashed_string = ascii_string.replace(" ", "-")
        email = f"{dashed_string}-fake@openmined.org"
    if password is None:
        password = fake.password()

    return TestUser(name=name, email=email, password=password, role=role)


def user_exists(root_client, email: str) -> bool:
    users = root_client.api.services.user
    for user in users:
        if user.email == email:
            return True
    return False


class SMTPTestServer:
    def __init__(self, email_server, port=9025, ready_timeout=5):
        self.port = port
        self.hostname = "0.0.0.0"  # nosec: B104
        self.controller = None

        # Simple email handler class
        class SimpleHandler:
            async def handle_DATA(self, server, session, envelope):
                try:
                    print(f"> SMTPTestServer got an email for {envelope.rcpt_tos}")
                    email = Email(
                        email_from=envelope.mail_from,
                        email_to=envelope.rcpt_tos,
                        email_content=envelope.content.decode(
                            "utf-8", errors="replace"
                        ),
                    )
                    email_server.add_email_for_user(envelope.rcpt_tos[0], email)
                    email_server.save_emails()
                    return "250 Message accepted for delivery"
                except Exception as e:
                    print(f"> Error handling email: {e}")
                    return "550 Internal Server Error"

        try:
            self.handler = SimpleHandler()
            self.controller = Controller(
                self.handler,
                hostname=self.hostname,
                port=self.port,
                ready_timeout=ready_timeout,
            )
        except Exception as e:
            print(f"> Error initializing SMTPTestServer Controller: {e}")

    def start(self):
        self.controller.start()

    def stop(self):
        self.controller.stop()

    def __del__(self):
        if self.controller:
            self.stop()


class TimeoutError(Exception):
    pass


class Timeout:
    def __init__(self, timeout_duration):
        if timeout_duration > 60:
            raise ValueError("Timeout duration cannot exceed 60 seconds.")
        self.timeout_duration = timeout_duration

    def run_with_timeout(self, condition_func, *args, **kwargs):
        start_time = time.time()
        result = None

        while True:
            elapsed_time = time.time() - start_time
            if elapsed_time > self.timeout_duration:
                raise TimeoutError(
                    f"Function execution exceeded {self.timeout_duration} seconds."
                )

            # Check if the condition is met
            try:
                if condition_func():
                    print("Condition met, exiting early.")
                    break
            except Exception as e:
                print(f"Exception in target function: {e}")
                break  # Exit the loop if an exception occurs in the function
            time.sleep(1)

        return result


def get_email_server(reset=False, port=9025):
    email_server = EmailServer()
    if reset:
        email_server.reset_emails()
    for _ in range(5):
        try:
            smtp_server = SMTPTestServer(email_server, port=port)
            smtp_server.start()
            return email_server, smtp_server

        except TimeoutError:
            del smtp_server
            print("SMTP server timed out. Retrying...")
            continue
        except Exception as e:
            print(f"> Error starting SMTP server: {e}")
    raise Exception("Failed to start SMTP server in 5 attempts.")


def create_user(root_client, test_user):
    if not user_exists(root_client, test_user.email):
        fake = Faker()
        root_client.register(
            name=test_user.name,
            email=test_user.email,
            password=test_user.password,
            password_verify=test_user.password,
            institution=fake.company(),
            website=fake.url(),
        )
    else:
        print("User already exists", test_user)



# stdlib

# third party

# relative
from ...serde.serializable import serializable
from ...server.credentials import SyftVerifyKey
from ...store.db.stash import ObjectStash
from ...store.document_store_errors import NotFoundException
from ...store.document_store_errors import StashException
from ...types.result import as_result
from .project import Project


@serializable(canonical_name="ProjectSQLStash", version=1)
class ProjectStash(ObjectStash[Project]):
    @as_result(StashException)
    def get_all_for_verify_key(
        self, credentials: SyftVerifyKey, verify_key: SyftVerifyKey
    ) -> list[Project]:
        return self.get_all(
            credentials=credentials,
            filters={"user_verify_key": verify_key},
        ).unwrap()

    @as_result(StashException, NotFoundException)
    def get_by_name(self, credentials: SyftVerifyKey, project_name: str) -> Project:
        return self.get_one(
            credentials=credentials,
            filters={"name": project_name},
        ).unwrap()



# future
from __future__ import annotations

# relative
from ...serde.serializable import serializable
from ...server.credentials import SyftVerifyKey
from ...store.db.stash import ObjectStash
from ...store.document_store_errors import NotFoundException
from ...store.document_store_errors import StashException
from ...types.errors import SyftException
from ...types.result import as_result
from ...types.syft_object import SyftObject
from ...types.twin_object import TwinObject
from ...types.uid import UID
from .action_object import ActionObject
from .action_object import is_action_data_empty
from .action_permissions import ActionObjectEXECUTE
from .action_permissions import ActionObjectPermission
from .action_permissions import ActionObjectREAD
from .action_permissions import ActionObjectWRITE
from .action_permissions import StoragePermission


@serializable(canonical_name="ActionObjectSQLStore", version=1)
class ActionObjectStash(ObjectStash[ActionObject]):
    # We are storing ActionObject, Action, TwinObject
    allow_any_type = True

    @as_result(NotFoundException, SyftException)
    def get(
        self, uid: UID, credentials: SyftVerifyKey, has_permission: bool = False
    ) -> ActionObject:
        uid = uid.id  # We only need the UID from LineageID or UID
        # TODO remove and use get_by_uid instead
        return self.get_by_uid(
            credentials=credentials,
            uid=uid,
            has_permission=has_permission,
        ).unwrap()

    @as_result(NotFoundException, SyftException)
    def get_mock(self, credentials: SyftVerifyKey, uid: UID) -> SyftObject:
        uid = uid.id  # We only need the UID from LineageID or UID

        obj = self.get_by_uid(
            credentials=credentials, uid=uid, has_permission=True
        ).unwrap()
        if isinstance(obj, TwinObject) and not is_action_data_empty(obj.mock):
            return obj.mock
        raise NotFoundException(public_message=f"No mock found for object {uid}")

    @as_result(NotFoundException, SyftException)
    def get_pointer(
        self,
        uid: UID,
        credentials: SyftVerifyKey,
        server_uid: UID,
    ) -> SyftObject:
        uid = uid.id  # We only need the UID from LineageID or UID

        obj = self.get_by_uid(
            credentials=credentials, uid=uid, has_permission=True
        ).unwrap()
        has_permissions = self.has_permission(
            ActionObjectREAD(uid=uid, credentials=credentials)
        )

        if has_permissions:
            if isinstance(obj, TwinObject):
                return obj.private.syft_point_to(server_uid)
            return obj.syft_point_to(server_uid)  # type: ignore

        # if its a twin with a mock anyone can have this
        if isinstance(obj, TwinObject):
            return obj.mock.syft_point_to(server_uid)

        # finally worst case you get ActionDataEmpty so you can still trace
        return obj.as_empty().syft_point_to(server_uid)  # type: ignore

    @as_result(SyftException, StashException)
    def set_or_update(  # type: ignore
        self,
        uid: UID,
        credentials: SyftVerifyKey,
        syft_object: SyftObject,
        has_result_read_permission: bool = False,
        add_storage_permission: bool = True,
    ) -> UID:
        uid = uid.id  # We only need the UID from LineageID or UID

        if self.exists(credentials=credentials, uid=uid):
            permissions: list[ActionObjectPermission] = []
            if has_result_read_permission:
                permissions.append(ActionObjectREAD(uid=uid, credentials=credentials))
            else:
                permissions.extend(
                    [
                        ActionObjectWRITE(uid=uid, credentials=credentials),
                        ActionObjectEXECUTE(uid=uid, credentials=credentials),
                    ]
                )
            storage_permission = []
            if add_storage_permission:
                storage_permission.append(
                    StoragePermission(uid=uid, server_uid=self.server_uid)
                )

            self.update(
                credentials=credentials,
                obj=syft_object,
            ).unwrap()
            self.add_permissions(permissions).unwrap()
            self.add_storage_permissions(storage_permission).unwrap()
            return uid

        owner_credentials = (
            credentials if has_result_read_permission else self.root_verify_key
        )
        # if not has_result_read_permission
        # root takes owneship, but you can still write and execute
        super().set(
            credentials=owner_credentials,
            obj=syft_object,
            add_permissions=[
                ActionObjectWRITE(uid=uid, credentials=credentials),
                ActionObjectEXECUTE(uid=uid, credentials=credentials),
            ],
            add_storage_permission=add_storage_permission,
        ).unwrap()

        return uid

    def set(self, *args, **kwargs):  # type: ignore
        raise Exception("Use `ActionObjectStash.set_or_update` instead.")



# stdlib
from binascii import hexlify
import itertools
import logging
import sys
import threading
from threading import Event
from time import sleep
from typing import Any

# third party
import zmq
from zmq import LINGER

# relative
from ...serde.serializable import serializable
from ...serde.serialize import _serialize as serialize
from ...service.action.action_object import ActionObject
from ...service.context import AuthedServiceContext
from ...types.errors import SyftException
from ...types.result import as_result
from ...types.uid import UID
from ...util.util import get_queue_address
from ..service import AbstractService
from ..worker.worker_pool import ConsumerState
from ..worker.worker_stash import WorkerStash
from .base_queue import QueueProducer
from .queue_stash import ActionQueueItem
from .queue_stash import QueueStash
from .queue_stash import Status
from .zmq_common import HEARTBEAT_INTERVAL_SEC
from .zmq_common import Service
from .zmq_common import THREAD_TIMEOUT_SEC
from .zmq_common import Timeout
from .zmq_common import Worker
from .zmq_common import ZMQCommand
from .zmq_common import ZMQHeader
from .zmq_common import ZMQ_POLLER_TIMEOUT_MSEC
from .zmq_common import ZMQ_SOCKET_LOCK

logger = logging.getLogger(__name__)


@serializable(canonical_name="ZMQProducer", version=1)
class ZMQProducer(QueueProducer):
    INTERNAL_SERVICE_PREFIX = b"mmi."

    def __init__(
        self,
        queue_name: str,
        queue_stash: QueueStash,
        worker_stash: WorkerStash,
        port: int,
        context: AuthedServiceContext,
    ) -> None:
        self.id = UID().short()
        self.port = port
        self.queue_stash = queue_stash
        self.worker_stash = worker_stash
        self.queue_name = queue_name
        self.auth_context = context
        self._stop = Event()
        self.post_init()

    @property
    def address(self) -> str:
        return get_queue_address(self.port)

    def post_init(self) -> None:
        """Initialize producer state."""

        self.services: dict[str, Service] = {}
        self.workers: dict[bytes, Worker] = {}
        self.waiting: list[Worker] = []
        self.heartbeat_t = Timeout(HEARTBEAT_INTERVAL_SEC)
        self.context = zmq.Context(1)
        self.socket = self.context.socket(zmq.ROUTER)
        self.socket.setsockopt(LINGER, 1)
        self.socket.setsockopt_string(zmq.IDENTITY, self.id)
        self.poll_workers = zmq.Poller()
        self.poll_workers.register(self.socket, zmq.POLLIN)
        self.bind(f"tcp://*:{self.port}")
        self.thread: threading.Thread | None = None
        self.producer_thread: threading.Thread | None = None

    def close(self) -> None:
        self._stop.set()
        try:
            if self.thread:
                self.thread.join(THREAD_TIMEOUT_SEC)
                if self.thread.is_alive():
                    logger.error(
                        f"ZMQProducer message sending thread join timed out during closing. "
                        f"Queue name {self.queue_name}, "
                    )
                self.thread = None

            if self.producer_thread:
                self.producer_thread.join(THREAD_TIMEOUT_SEC)
                if self.producer_thread.is_alive():
                    logger.error(
                        f"ZMQProducer queue thread join timed out during closing. "
                        f"Queue name {self.queue_name}, "
                    )
                self.producer_thread = None

            self.poll_workers.unregister(self.socket)
        except Exception as e:
            logger.exception("Failed to unregister poller.", exc_info=e)
        finally:
            self.socket.close()
            self.context.destroy()

    @property
    def action_service(self) -> AbstractService:
        if self.auth_context.server is not None:
            return self.auth_context.server.services.action
        else:
            raise Exception(f"{self.auth_context} does not have a server.")

    @as_result(SyftException)
    def contains_unresolved_action_objects(self, arg: Any, recursion: int = 0) -> bool:
        """recursively check collections for unresolved action objects"""
        if isinstance(arg, UID):
            arg = self.action_service.get(self.auth_context, arg)
            return self.contains_unresolved_action_objects(
                arg, recursion=recursion + 1
            ).unwrap()
        if isinstance(arg, ActionObject):
            if not arg.syft_resolved:
                arg = self.action_service.get(self.auth_context, arg)
                if not arg.syft_resolved:
                    return True
            arg = arg.syft_action_data

        value = False
        if isinstance(arg, list):
            for elem in arg:
                value = self.contains_unresolved_action_objects(
                    elem, recursion=recursion + 1
                ).unwrap()
                if value:
                    return True
        if isinstance(arg, dict):
            for elem in arg.values():
                value = self.contains_unresolved_action_objects(
                    elem, recursion=recursion + 1
                ).unwrap()
                if value:
                    return True
        return value

    def read_items(self) -> None:
        while True:
            if self._stop.is_set():
                break
            try:
                sleep(1)

                # Items to be queued
                items_to_queue = self.queue_stash.get_by_status(
                    self.queue_stash.root_verify_key,
                    status=Status.CREATED,
                ).unwrap()

                items_to_queue = [] if items_to_queue is None else items_to_queue

                # Queue Items that are in the processing state
                items_processing = self.queue_stash.get_by_status(
                    self.queue_stash.root_verify_key,
                    status=Status.PROCESSING,
                ).unwrap()

                items_processing = [] if items_processing is None else items_processing

                for item in itertools.chain(items_to_queue, items_processing):
                    # TODO: if resolving fails, set queueitem to errored, and jobitem as well
                    if item.status == Status.CREATED:
                        if isinstance(item, ActionQueueItem):
                            action = item.kwargs["action"]
                            if (
                                self.contains_unresolved_action_objects(
                                    action.args
                                ).unwrap()
                                or self.contains_unresolved_action_objects(
                                    action.kwargs
                                ).unwrap()
                            ):
                                continue

                        msg_bytes = serialize(item, to_bytes=True)
                        worker_pool = item.worker_pool.resolve_with_context(
                            self.auth_context
                        ).unwrap()
                        service_name = worker_pool.name
                        service: Service | None = self.services.get(service_name)

                        # Skip adding message if corresponding service/pool
                        # is not registered.
                        if service is None:
                            continue

                        # append request message to the corresponding service
                        # This list is processed in dispatch method.

                        # TODO: Logic to evaluate the CAN RUN Condition
                        item.status = Status.PROCESSING
                        self.queue_stash.update(
                            item.syft_client_verify_key, item
                        ).unwrap(public_message=f"failed to update queue item {item}")
                        service.requests.append(msg_bytes)
                    elif item.status == Status.PROCESSING:
                        # Evaluate Retry condition here
                        # If job running and timeout or job status is KILL
                        # or heartbeat fails
                        # or container id doesn't exists, kill process or container
                        # else decrease retry count and mark status as CREATED.
                        pass
            except Exception as e:
                # stdlib
                import traceback

                print(e, traceback.format_exc(), file=sys.stderr)
                item.status = Status.ERRORED
                self.queue_stash.update(item.syft_client_verify_key, item).unwrap()

    def run(self) -> None:
        self.thread = threading.Thread(target=self._run)
        self.thread.start()

        self.producer_thread = threading.Thread(target=self.read_items)
        self.producer_thread.start()

    def send(self, worker: bytes, message: bytes | list[bytes]) -> None:
        worker_obj = self.require_worker(worker)
        self.send_to_worker(worker_obj, ZMQCommand.W_REQUEST, message)

    def bind(self, endpoint: str) -> None:
        """Bind producer to endpoint."""
        self.socket.bind(endpoint)
        logger.info(f"ZMQProducer endpoint: {endpoint}")

    def send_heartbeats(self) -> None:
        """Send heartbeats to idle workers if it's time"""
        if self.heartbeat_t.has_expired():
            for worker in self.waiting:
                self.send_to_worker(worker, ZMQCommand.W_HEARTBEAT)
            self.heartbeat_t.reset()

    def purge_workers(self) -> None:
        """Look for & kill expired workers.

        Workers are oldest to most recent, so we stop at the first alive worker.
        """
        # work on a copy of the iterator
        for worker in self.waiting:
            res = worker._syft_worker(self.worker_stash, self.auth_context.credentials)
            if res.is_err() or (syft_worker := res.ok()) is None:
                logger.info(f"Failed to retrieve SyftWorker {worker.syft_worker_id}")
                continue

            if worker.has_expired() or syft_worker.to_be_deleted:
                logger.info(f"Deleting expired worker id={worker}")
                self.delete_worker(worker, syft_worker.to_be_deleted)

                # relative

                self.auth_context.server.services.worker._delete(
                    self.auth_context, syft_worker
                )

    def update_consumer_state_for_worker(
        self, syft_worker_id: UID, consumer_state: ConsumerState
    ) -> None:
        if self.worker_stash is None:
            logger.error(  # type: ignore[unreachable]
                f"ZMQProducer worker stash not defined for {self.queue_name} - {self.id}"
            )
            return

        try:
            try:
                self.worker_stash.get_by_uid(
                    credentials=self.worker_stash.root_verify_key,
                    uid=syft_worker_id,
                ).unwrap()
            except Exception:
                return None

            self.worker_stash.update_consumer_state(
                credentials=self.worker_stash.root_verify_key,
                worker_uid=syft_worker_id,
                consumer_state=consumer_state,
            ).unwrap()
        except Exception:
            logger.exception(
                f"Failed to update consumer state for worker id: {syft_worker_id} to state {consumer_state}",
            )

    def worker_waiting(self, worker: Worker) -> None:
        """This worker is now waiting for work."""
        # Queue to broker and service waiting lists
        if worker not in self.waiting:
            self.waiting.append(worker)
        if worker.service is not None and worker not in worker.service.waiting:
            worker.service.waiting.append(worker)
        worker.reset_expiry()
        self.update_consumer_state_for_worker(worker.syft_worker_id, ConsumerState.IDLE)
        self.dispatch(worker.service, None)

    def dispatch(self, service: Service, msg: bytes) -> None:
        """Dispatch requests to waiting workers as possible"""
        if msg is not None:  # Queue message if any
            service.requests.append(msg)

        self.purge_workers()
        while service.waiting and service.requests:
            # One worker consuming only one message at a time.
            msg = service.requests.pop(0)
            worker = service.waiting.pop(0)
            self.waiting.remove(worker)
            self.send_to_worker(worker, ZMQCommand.W_REQUEST, msg)

    def send_to_worker(
        self,
        worker: Worker,
        command: bytes,
        msg: bytes | list | None = None,
    ) -> None:
        """Send message to worker.

        If message is provided, sends that message.
        """

        if self.socket.closed:
            logger.warning("Socket is closed. Cannot send message.")
            return

        if msg is None:
            msg = []
        elif not isinstance(msg, list):
            msg = [msg]

        # ZMQProducer send frames: [address, empty, header, command, ...data]
        core = [worker.address, b"", ZMQHeader.W_WORKER, command]
        msg = core + msg

        if command != ZMQCommand.W_HEARTBEAT:
            # log everything except the last frame which contains serialized data
            logger.info(f"ZMQProducer send: {core}")

        with ZMQ_SOCKET_LOCK:
            try:
                self.socket.send_multipart(msg)
            except zmq.ZMQError:
                logger.exception("ZMQProducer send error")

    def _run(self) -> None:
        try:
            while True:
                if self._stop.is_set():
                    logger.info("ZMQProducer thread stopped")
                    return

                for service in self.services.values():
                    self.dispatch(service, None)

                items = None

                try:
                    items = self.poll_workers.poll(ZMQ_POLLER_TIMEOUT_MSEC)
                except Exception as e:
                    logger.exception("ZMQProducer poll error", exc_info=e)

                if items:
                    msg = self.socket.recv_multipart()

                    if len(msg) < 3:
                        logger.error(f"ZMQProducer invalid recv: {msg}")
                        continue

                    # ZMQProducer recv frames: [address, empty, header, command, ...data]
                    (address, _, header, command, *data) = msg

                    if command != ZMQCommand.W_HEARTBEAT:
                        # log everything except the last frame which contains serialized data
                        logger.info(f"ZMQProducer recv: {msg[:4]}")

                    if header == ZMQHeader.W_WORKER:
                        self.process_worker(address, command, data)
                    else:
                        logger.error(f"Invalid message header: {header}")

                self.send_heartbeats()
                self.purge_workers()
        except Exception as e:
            logger.exception("ZMQProducer thread exception", exc_info=e)

    def require_worker(self, address: bytes) -> Worker:
        """Finds the worker (creates if necessary)."""
        identity = hexlify(address)
        worker = self.workers.get(identity)
        if worker is None:
            worker = Worker(identity=identity, address=address)
            self.workers[identity] = worker
        return worker

    def process_worker(self, address: bytes, command: bytes, data: list[bytes]) -> None:
        worker_ready = hexlify(address) in self.workers
        worker = self.require_worker(address)

        if ZMQCommand.W_READY == command:
            service_name = data.pop(0).decode()
            syft_worker_id = data.pop(0).decode()
            if worker_ready:
                # Not first command in session or Reserved service name
                # If worker was already present, then we disconnect it first
                # and wait for it to re-register itself to the producer. This ensures that
                # we always have a healthy worker in place that can talk to the producer.
                self.delete_worker(worker, True)
            else:
                # Attach worker to service and mark as idle
                if service_name in self.services:
                    service: Service | None = self.services.get(service_name)
                else:
                    service = Service(service_name)
                    self.services[service_name] = service
                if service is not None:
                    worker.service = service
                logger.info(f"New worker: {worker}")
                worker.syft_worker_id = UID(syft_worker_id)
                self.worker_waiting(worker)

        elif ZMQCommand.W_HEARTBEAT == command:
            if worker_ready:
                # If worker is ready then reset expiry
                # and add it to worker waiting list
                # if not already present
                self.worker_waiting(worker)
            else:
                logger.info(f"Got heartbeat, but worker not ready. {worker}")
                self.delete_worker(worker, True)
        elif ZMQCommand.W_DISCONNECT == command:
            logger.info(f"Removing disconnected worker: {worker}")
            self.delete_worker(worker, False)
        else:
            logger.error(f"Invalid command: {command!r}")

    def delete_worker(self, worker: Worker, disconnect: bool) -> None:
        """Deletes worker from all data structures, and deletes worker."""
        if disconnect:
            self.send_to_worker(worker, ZMQCommand.W_DISCONNECT)

        if worker.service and worker in worker.service.waiting:
            worker.service.waiting.remove(worker)

        if worker in self.waiting:
            self.waiting.remove(worker)

        self.workers.pop(worker.identity, None)

        if worker.syft_worker_id is not None:
            self.update_consumer_state_for_worker(
                worker.syft_worker_id, ConsumerState.DETACHED
            )

    @property
    def alive(self) -> bool:
        return not self.socket.closed



# stdlib
import logging
import subprocess  # nosec
import threading
from threading import Event

# third party
import zmq
from zmq import Frame
from zmq.error import ContextTerminated

# relative
from ...serde.deserialize import _deserialize
from ...serde.serializable import serializable
from ...types.uid import UID
from ..worker.worker_pool import ConsumerState
from ..worker.worker_stash import WorkerStash
from .base_queue import AbstractMessageHandler
from .base_queue import QueueConsumer
from .zmq_common import HEARTBEAT_INTERVAL_SEC
from .zmq_common import PRODUCER_TIMEOUT_SEC
from .zmq_common import THREAD_TIMEOUT_SEC
from .zmq_common import Timeout
from .zmq_common import ZMQCommand
from .zmq_common import ZMQHeader
from .zmq_common import ZMQ_POLLER_TIMEOUT_MSEC
from .zmq_common import ZMQ_SOCKET_LOCK

logger = logging.getLogger(__name__)


def last_created_port() -> int:
    command = (
        "lsof -i -P -n | grep '*:[0-9]* (LISTEN)' | grep python | awk '{print $9, $1, $2}' | "
        "sort -k2,2 -k3,3n | tail -n 1 | awk '{print $1}' | cut -d':' -f2"
    )
    # 1. Lists open files (including network connections) with lsof -i -P -n
    # 2. Filters for listening ports with grep '*:[0-9]* (LISTEN)'
    # 3. Further filters for Python processes with grep python
    # 4. Sorts based on the 9th field (which is likely the port number) with sort -k9
    # 5. Takes the last 10 entries with tail -n 10
    # 6. Prints only the 9th field (port and address) with awk '{print $9}'
    # 7. Extracts only the port number with cut -d':' -f2

    process = subprocess.Popen(  # nosec
        command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True
    )
    out, err = process.communicate()
    return int(out.decode("utf-8").strip())


@serializable(attrs=["_subscriber"], canonical_name="ZMQConsumer", version=1)
class ZMQConsumer(QueueConsumer):
    def __init__(
        self,
        message_handler: AbstractMessageHandler,
        address: str,
        queue_name: str,
        service_name: str,
        syft_worker_id: UID | None = None,
        worker_stash: WorkerStash | None = None,
        verbose: bool = False,
    ) -> None:
        self.address = address
        self.message_handler = message_handler
        self.service_name = service_name
        self.queue_name = queue_name
        self.context = zmq.Context()
        self.poller = zmq.Poller()
        self.socket = None
        self.verbose = verbose
        self.id = UID().short()
        self._stop = Event()
        self.syft_worker_id = syft_worker_id
        self.worker_stash = worker_stash
        self.post_init()

    @classmethod
    def default(cls, address: str | None = None, **kwargs: dict) -> "ZMQConsumer":
        # relative
        from ...types.uid import UID
        from ..worker.utils import DEFAULT_WORKER_POOL_NAME
        from .queue import APICallMessageHandler

        if address is None:
            try:
                address = f"tcp://localhost:{last_created_port()}"
            except Exception:
                raise Exception(
                    "Could not auto-assign ZMQConsumer address. Please provide one."
                )
            print(f"Auto-assigning ZMQConsumer address: {address}. Please verify.")
        default_kwargs = {
            "message_handler": APICallMessageHandler,
            "queue_name": APICallMessageHandler.queue_name,
            "service_name": DEFAULT_WORKER_POOL_NAME,
            "syft_worker_id": UID(),
            "verbose": True,
            "address": address,
        }

        for key, value in kwargs.items():
            if key in default_kwargs:
                default_kwargs[key] = value

        return cls(**default_kwargs)

    def reconnect_to_producer(self) -> None:
        """Connect or reconnect to producer"""
        if self.socket:
            self.poller.unregister(self.socket)  # type: ignore[unreachable]
            self.socket.close()
        self.socket = self.context.socket(zmq.DEALER)
        self.socket.linger = 0
        self.socket.setsockopt_string(zmq.IDENTITY, self.id)
        self.socket.connect(self.address)
        self.poller.register(self.socket, zmq.POLLIN)

        logger.info(f"Connecting Worker id={self.id} to broker addr={self.address}")

        # Register queue with the producer
        self.send_to_producer(
            ZMQCommand.W_READY,
            [self.service_name.encode(), str(self.syft_worker_id).encode()],
        )

    def post_init(self) -> None:
        self.thread: threading.Thread | None = None
        self.heartbeat_t = Timeout(HEARTBEAT_INTERVAL_SEC)
        self.producer_ping_t = Timeout(PRODUCER_TIMEOUT_SEC)
        self.reconnect_to_producer()

    def disconnect_from_producer(self) -> None:
        self.send_to_producer(ZMQCommand.W_DISCONNECT)

    def close(self) -> None:
        self.disconnect_from_producer()
        self._stop.set()
        try:
            if self.thread is not None:
                self.thread.join(timeout=THREAD_TIMEOUT_SEC)
                if self.thread is not None and self.thread.is_alive():
                    logger.error(
                        f"ZMQConsumer thread join timed out during closing. "
                        f"SyftWorker id {self.syft_worker_id}, "
                        f"service name {self.service_name}."
                    )
                self.thread = None
            self.poller.unregister(self.socket)
        except Exception as e:
            logger.error("Failed to unregister worker.", exc_info=e)
        finally:
            self.socket.close()
            self.context.destroy()
            # self._stop.clear()

    def send_to_producer(
        self,
        command: bytes,
        msg: bytes | list | None = None,
    ) -> None:
        """Send message to producer.

        If no msg is provided, creates one internally
        """
        if self.socket.closed:
            logger.warning("Socket is closed. Cannot send message.")
            return

        if msg is None:
            msg = []
        elif not isinstance(msg, list):
            msg = [msg]

        # ZMQConsumer send frames: [empty, header, command, ...data]
        core = [b"", ZMQHeader.W_WORKER, command]
        msg = core + msg

        if command != ZMQCommand.W_HEARTBEAT:
            logger.info(f"ZMQ Consumer send: {core}")

        with ZMQ_SOCKET_LOCK:
            try:
                self.socket.send_multipart(msg)
            except zmq.ZMQError as e:
                logger.error("ZMQConsumer send error", exc_info=e)

    def _run(self) -> None:
        """Send reply, if any, to producer and wait for next request."""
        try:
            while True:
                if self._stop.is_set():
                    logger.info("ZMQConsumer thread stopped")
                    return

                try:
                    items = self.poller.poll(ZMQ_POLLER_TIMEOUT_MSEC)
                except ContextTerminated:
                    logger.info("Context terminated")
                    return
                except Exception as e:
                    logger.error("ZMQ poll error", exc_info=e)
                    continue

                if items:
                    msg = self.socket.recv_multipart()

                    # mark as alive
                    self.set_producer_alive()

                    if len(msg) < 3:
                        logger.error(f"ZMQConsumer invalid recv: {msg}")
                        continue

                    # Message frames recieved by consumer:
                    # [empty, header, command, ...data]
                    (_, _, command, *data) = msg

                    if command != ZMQCommand.W_HEARTBEAT:
                        # log everything except the last frame which contains serialized data
                        logger.info(f"ZMQConsumer recv: {msg[:-4]}")

                    if command == ZMQCommand.W_REQUEST:
                        # Call Message Handler
                        try:
                            message = data.pop()
                            self.associate_job(message)
                            self.message_handler.handle_message(
                                message=message,
                                syft_worker_id=self.syft_worker_id,
                            )
                        except Exception as e:
                            logger.exception("Couldn't handle message", exc_info=e)
                        finally:
                            self.clear_job()
                    elif command == ZMQCommand.W_HEARTBEAT:
                        self.set_producer_alive()
                    elif command == ZMQCommand.W_DISCONNECT:
                        self.reconnect_to_producer()
                    else:
                        logger.error(f"ZMQConsumer invalid command: {command}")
                else:
                    if not self.is_producer_alive():
                        logger.info("Producer check-alive timed out. Reconnecting.")
                        self.reconnect_to_producer()
                        self.set_producer_alive()

                if not self._stop.is_set():
                    self.send_heartbeat()

        except zmq.ZMQError as e:
            if e.errno == zmq.ETERM:
                logger.info("zmq.ETERM")
            else:
                logger.exception("zmq.ZMQError", exc_info=e)
        except Exception as e:
            logger.exception("ZMQConsumer thread exception", exc_info=e)

    def set_producer_alive(self) -> None:
        self.producer_ping_t.reset()

    def is_producer_alive(self) -> bool:
        # producer timer is within timeout
        return not self.producer_ping_t.has_expired()

    def send_heartbeat(self) -> None:
        if self.heartbeat_t.has_expired() and self.is_producer_alive():
            self.send_to_producer(ZMQCommand.W_HEARTBEAT)
            self.heartbeat_t.reset()

    def run(self) -> None:
        self.thread = threading.Thread(target=self._run)
        self.thread.start()

    def associate_job(self, message: Frame) -> None:
        try:
            queue_item = _deserialize(message, from_bytes=True)
            self._set_worker_job(queue_item.job_id)
        except Exception as e:
            logger.exception("Could not associate job", exc_info=e)

    def clear_job(self) -> None:
        self._set_worker_job(None)

    def _set_worker_job(self, job_id: UID | None) -> None:
        if self.worker_stash is not None:
            consumer_state = (
                ConsumerState.IDLE if job_id is None else ConsumerState.CONSUMING
            )
            res = self.worker_stash.update_consumer_state(
                credentials=self.worker_stash.root_verify_key,
                worker_uid=self.syft_worker_id,
                consumer_state=consumer_state,
            )
            if res.is_err():
                logger.error(
                    f"Failed to update consumer state for {self.service_name}-{self.id}, error={res.err()}"
                )

    @property
    def alive(self) -> bool:
        return not self.socket.closed and self.is_producer_alive()



def make_nb_logger(name="notebook"):
    # stdlib
    import logging

    logger = logging.getLogger()
    file_logging = logging.FileHandler(f"{name}.log")
    file_logging.setFormatter(
        logging.Formatter(
            "%(asctime)s - pid-%(process)d tid-%(thread)d - %(levelname)s - %(name)s - %(message)s"
        )
    )
    file_logging.setLevel(logging.INFO)

    logger.addHandler(file_logging)
    logger.setLevel(logging.INFO)
    return logger



# stdlib

# relative
from ...serde.serializable import serializable
from ...server.credentials import SyftVerifyKey
from ...store.db.db import DBManager
from ...store.db.stash import ObjectStash
from ...store.document_store_errors import StashException
from ...types.result import as_result
from ..context import AuthedServiceContext
from ..response import SyftSuccess
from ..service import AbstractService
from ..service import SERVICE_TO_TYPES
from ..service import TYPE_TO_SERVICE
from .data_subject_member import DataSubjectMemberRelationship


@serializable(canonical_name="DataSubjectMemberSQLStash", version=1)
class DataSubjectMemberStash(ObjectStash[DataSubjectMemberRelationship]):
    @as_result(StashException)
    def get_all_for_parent(
        self, credentials: SyftVerifyKey, name: str
    ) -> list[DataSubjectMemberRelationship]:
        return self.get_all(
            credentials=credentials,
            filters={"parent": name},
        ).unwrap()

    @as_result(StashException)
    def get_all_for_child(
        self, credentials: SyftVerifyKey, name: str
    ) -> list[DataSubjectMemberRelationship]:
        return self.get_all(
            credentials=credentials,
            filters={"child": name},
        ).unwrap()


@serializable(canonical_name="DataSubjectMemberService", version=1)
class DataSubjectMemberService(AbstractService):
    stash: DataSubjectMemberStash

    def __init__(self, store: DBManager) -> None:
        self.stash = DataSubjectMemberStash(store=store)

    def add(
        self, context: AuthedServiceContext, parent: str, child: str
    ) -> SyftSuccess:
        """Register relationship between data subject and it's member."""
        relation = DataSubjectMemberRelationship(parent=parent, child=child)
        self.stash.set(context.credentials, relation, ignore_duplicates=True).unwrap()
        return SyftSuccess(message=f"Relationship added for: {parent} -> {child}")

    def get_relatives(
        self, context: AuthedServiceContext, data_subject_name: str
    ) -> list[DataSubjectMemberRelationship]:
        """Get all Members for given data subject"""
        return self.stash.get_all_for_parent(
            context.credentials, name=data_subject_name
        ).unwrap()


TYPE_TO_SERVICE[DataSubjectMemberRelationship] = DataSubjectMemberService
SERVICE_TO_TYPES[DataSubjectMemberService].update({DataSubjectMemberRelationship})



# stdlib
import logging

# third party
from sqlalchemy.exc import DatabaseError
from typing_extensions import Self

# relative
from ..document_store_errors import StashException

logger = logging.getLogger(__name__)


class StashDBException(StashException):
    """
    See https://docs.sqlalchemy.org/en/20/errors.html#databaseerror

    StashDBException converts a SQLAlchemy DatabaseError into a StashException,
    DatabaseErrors are errors thrown by the database itself, for example when a
    query fails because a table is missing.
    """

    public_message = "There was an error retrieving data. Contact your admin."

    @classmethod
    def from_sqlalchemy_error(cls, e: DatabaseError) -> Self:
        logger.exception(e)

        error_type = e.__class__.__name__
        private_message = f"{error_type}: {str(e)}"
        return cls(private_message=private_message)



# stdlib
from collections import defaultdict
import logging
from typing import Any

# syft absolute
import syft

# relative
from ...serde.serializable import serializable
from ...store.db.db import DBManager
from ...store.db.stash import ObjectStash
from ...store.document_store_errors import NotFoundException
from ...types.blob_storage import BlobStorageEntry
from ...types.errors import SyftException
from ...types.result import as_result
from ...types.syft_object import SyftObject
from ...types.syft_object_registry import SyftObjectRegistry
from ...types.twin_object import TwinObject
from ...types.uid import UID
from ..action.action_object import Action
from ..action.action_object import ActionObject
from ..action.action_permissions import ActionObjectPermission
from ..action.action_permissions import StoragePermission
from ..action.action_store import ActionObjectStash
from ..context import AuthedServiceContext
from ..response import SyftError
from ..response import SyftSuccess
from ..service import AbstractService
from ..service import service_method
from ..sync.sync_service import get_store
from ..sync.sync_service import get_store_by_type
from ..user.user_roles import ADMIN_ROLE_LEVEL
from ..user.user_roles import DATA_SCIENTIST_ROLE_LEVEL
from ..worker.utils import DEFAULT_WORKER_POOL_NAME
from .object_migration_state import MigrationData
from .object_migration_state import StoreMetadata
from .object_migration_state import SyftMigrationStateStash
from .object_migration_state import SyftObjectMigrationState

logger = logging.getLogger(__name__)


@serializable(canonical_name="MigrationService", version=1)
class MigrationService(AbstractService):
    stash: SyftMigrationStateStash

    def __init__(self, store: DBManager) -> None:
        self.stash = SyftMigrationStateStash(store=store)

    @service_method(path="migration", name="get_version")
    def get_version(self, context: AuthedServiceContext, canonical_name: str) -> int:
        """Search for the metadata for an object."""

        migration_state = self.stash.get_by_name(
            canonical_name=canonical_name, credentials=context.credentials
        ).unwrap()

        if migration_state is None:
            raise SyftException(
                public_message=f"No migration state exists for canonical name: {canonical_name}"
            )

        return migration_state.current_version

    @service_method(path="migration", name="get_state")
    @as_result(SyftException, NotFoundException)
    def get_state(
        self, context: AuthedServiceContext, canonical_name: str
    ) -> SyftObjectMigrationState:
        return self.stash.get_by_name(
            canonical_name=canonical_name, credentials=context.credentials
        ).unwrap()

    @service_method(path="migration", name="register_migration_state")
    def register_migration_state(
        self,
        context: AuthedServiceContext,
        current_version: int,
        canonical_name: str,
    ) -> SyftObjectMigrationState:
        obj = SyftObjectMigrationState(
            current_version=current_version, canonical_name=canonical_name
        )
        return self.stash.set(obj=obj, credentials=context.credentials).unwrap()

    @as_result(SyftException, NotFoundException)
    def _find_klasses_pending_for_migration(
        self, context: AuthedServiceContext, object_types: list[type[SyftObject]]
    ) -> list[type[SyftObject]]:
        klasses_to_be_migrated = []

        for object_type in object_types:
            canonical_name = object_type.__canonical_name__
            object_version = object_type.__version__

            try:
                migration_state = self.get_state(context, canonical_name).unwrap(
                    public_message=f"Failed to get migration state for {canonical_name}."
                )
                if int(migration_state.current_version) != int(
                    migration_state.latest_version
                ):
                    klasses_to_be_migrated.append(object_type)
            except NotFoundException:
                self.register_migration_state(
                    context,
                    current_version=object_version,
                    canonical_name=canonical_name,
                )

        return klasses_to_be_migrated

    @service_method(
        path="migration.get_all_store_metadata",
        name="get_all_store_metadata",
        roles=ADMIN_ROLE_LEVEL,
    )
    def get_all_store_metadata(
        self,
        context: AuthedServiceContext,
        document_store_object_types: list[type[SyftObject]] | None = None,
        include_action_store: bool = True,
    ) -> dict[type[SyftObject], StoreMetadata]:
        return self._get_all_store_metadata(
            context,
            document_store_object_types=document_store_object_types,
        ).unwrap()

    @as_result(SyftException)
    def _get_all_store_metadata(
        self,
        context: AuthedServiceContext,
        document_store_object_types: list[type[SyftObject]] | None = None,
    ) -> dict[type[SyftObject], StoreMetadata]:
        # metadata = permissions + storage permissions
        stashes = context.server.services.stashes
        store_metadata = {}

        for klass, stash in stashes.items():
            if (
                document_store_object_types is not None
                and klass not in document_store_object_types
            ):
                continue
            store_metadata[klass] = StoreMetadata(
                object_type=klass,
                permissions=stash.get_all_permissions().unwrap(),
                storage_permissions=stash.get_all_storage_permissions().unwrap(),
            )

        return store_metadata

    @as_result(SyftException)
    def _update_store_metadata_for_klass(
        self, context: AuthedServiceContext, metadata: StoreMetadata
    ) -> None:
        stash = self._search_stash_for_klass(context, metadata.object_type).unwrap()
        permissions = [
            ActionObjectPermission.from_permission_string(uid, perm_str)
            for uid, perm_strs in metadata.permissions.items()
            for perm_str in perm_strs
        ]

        storage_permissions = [
            StoragePermission(uid, server_uid)
            for uid, server_uids in metadata.storage_permissions.items()
            for server_uid in server_uids
        ]

        stash.add_permissions(permissions, ignore_missing=True).unwrap()
        stash.add_storage_permissions(storage_permissions, ignore_missing=True).unwrap()

    @as_result(SyftException)
    def _update_store_metadata(
        self, context: AuthedServiceContext, store_metadata: dict[type, StoreMetadata]
    ) -> None:
        print("Updating store metadata")
        for metadata in store_metadata.values():
            self._update_store_metadata_for_klass(context, metadata).unwrap()

    @as_result(SyftException)
    def _get_migration_objects(
        self,
        context: AuthedServiceContext,
        document_store_object_types: list[type[SyftObject]] | None = None,
        get_all: bool = False,
    ) -> dict[type[SyftObject], list[SyftObject]]:
        if document_store_object_types is None:
            document_store_object_types = list(context.server.services.stashes.keys())

        if get_all:
            klasses_to_migrate = document_store_object_types
        else:
            klasses_to_migrate = self._find_klasses_pending_for_migration(
                context=context, object_types=document_store_object_types
            ).unwrap()

        result = defaultdict(list)

        for klass in klasses_to_migrate:
            stash_or_err = self._search_stash_for_klass(context, klass)
            if stash_or_err.is_err():
                continue
            stash = stash_or_err.unwrap()

            for object in stash._data:
                actual_klass = type(object)
                use_klass = (
                    klass
                    if actual_klass.__canonical_name__ == klass.__canonical_name__
                    else actual_klass
                )
                result[use_klass].append(object)

        return dict(result)

    @as_result(SyftException)
    def _search_stash_for_klass(
        self, context: AuthedServiceContext, klass: type[SyftObject]
    ) -> ObjectStash:
        if issubclass(klass, ActionObject | TwinObject | Action):
            return context.server.services.action.stash

        stashes: dict[str, ObjectStash] = {  # type: ignore
            t.__canonical_name__: stash
            for t, stash in context.server.services.stashes.items()
        }

        mro = klass.__mro__
        class_index = 0
        object_stash = None
        while len(mro) > class_index:
            try:
                canonical_name = mro[class_index].__canonical_name__
            except AttributeError:
                # Classes without cname dont have a stash
                break
            object_stash = stashes.get(canonical_name)
            if object_stash is not None:
                break
            class_index += 1
        if object_stash is None:
            raise SyftException(public_message=f"Object stash not found for {klass}")
        return object_stash

    @service_method(
        path="migration.create_migrated_objects",
        name="create_migrated_objects",
        roles=ADMIN_ROLE_LEVEL,
    )
    def create_migrated_objects(
        self,
        context: AuthedServiceContext,
        migrated_objects: list[SyftObject],
        ignore_existing: bool = True,
    ) -> SyftSuccess:
        self._create_migrated_objects(
            context, migrated_objects, ignore_existing=ignore_existing
        ).unwrap()
        return SyftSuccess(message="Created migration objects!")

    @as_result(SyftException)
    def _create_migrated_objects(
        self,
        context: AuthedServiceContext,
        migrated_objects: dict[type[SyftObject], list[SyftObject]],
        ignore_existing: bool = True,
        skip_check_type: bool = False,
    ) -> dict[type[SyftObject], list[SyftObject]]:
        created_objects: dict[type[SyftObject], list[SyftObject]] = {}

        for key, objects in migrated_objects.items():
            created_objects[key] = []
            for migrated_object in objects:
                stash = self._search_stash_for_klass(
                    context, type(migrated_object)
                ).unwrap()

                result = stash.set(
                    context.credentials,
                    obj=migrated_object,
                    skip_check_type=skip_check_type,
                )
                # Exception from the new Error Handling pattern, no need to change
                if result.is_err():
                    # TODO: subclass a DuplicationKeyError
                    if ignore_existing and (
                        "Duplication Key Error" in result.err()._private_message  # type: ignore
                        or "Duplication Key Error" in result.err().public_message  # type: ignore
                    ):
                        print(
                            f"{type(migrated_object)} #{migrated_object.id} already exists"
                        )
                        continue
                    else:
                        result.unwrap()  # this will raise the exception inside the wrapper
                created_objects[key].append(result.unwrap())
        return created_objects

    @as_result(SyftException)
    def _update_migrated_objects(
        self, context: AuthedServiceContext, migrated_objects: list[SyftObject]
    ) -> SyftSuccess:
        for migrated_object in migrated_objects:
            stash = self._search_stash_for_klass(
                context, type(migrated_object)
            ).unwrap()

            stash.update(
                context.credentials,
                obj=migrated_object,
            ).unwrap()

        return SyftSuccess(message="Updated migration objects!")

    @as_result(SyftException)
    def _migrate_objects(
        self,
        context: AuthedServiceContext,
        migration_objects: dict[type[SyftObject], list[SyftObject]],
    ) -> list[SyftObject]:
        migrated_objects = []

        for klass, objects in migration_objects.items():
            canonical_name = klass.__canonical_name__
            latest_version = SyftObjectRegistry.get_latest_version(canonical_name)

            # Migrate data for objects in document store
            logger.info(
                f"Migrating data for: {canonical_name} table to version {latest_version}"
            )
            for object in objects:
                try:
                    migrated_value = object.migrate_to(latest_version, context)
                    migrated_objects.append(migrated_value)
                except Exception:
                    raise SyftException(
                        public_message=f"Failed to migrate data to {klass} for qk {klass.__version__}: {object.id}"
                    )
        return migrated_objects

    @service_method(
        path="migration.migrate_data",
        name="migrate_data",
        roles=ADMIN_ROLE_LEVEL,
    )
    def migrate_data(
        self,
        context: AuthedServiceContext,
        document_store_object_types: list[type[SyftObject]] | None = None,
    ) -> SyftSuccess:
        migration_objects = self._get_migration_objects(
            context, document_store_object_types
        ).unwrap()
        migrated_objects = self._migrate_objects(context, migration_objects).unwrap()
        self._update_migrated_objects(context, migrated_objects).unwrap()

        migration_actionobjects = self._get_migration_actionobjects(context).unwrap()
        migrated_actionobjects = self._migrate_objects(
            context, migration_actionobjects
        ).unwrap()
        self._update_migrated_actionobjects(context, migrated_actionobjects).unwrap()

        return SyftSuccess(message="Data upgraded to the latest version")

    @service_method(
        path="migration.get_migration_actionobjects",
        name="get_migration_actionobjects",
        roles=ADMIN_ROLE_LEVEL,
    )
    def get_migration_actionobjects(
        self, context: AuthedServiceContext, get_all: bool = False
    ) -> dict:
        return self._get_migration_actionobjects(context, get_all=get_all).unwrap()

    @as_result(SyftException)
    def _get_migration_actionobjects(
        self, context: AuthedServiceContext, get_all: bool = False
    ) -> dict[type[SyftObject], list[SyftObject]]:
        # Track all object types from action store
        action_object_types = [Action, ActionObject, TwinObject]
        action_object_types.extend(ActionObject.__subclasses__())
        klass_by_canonical_name: dict[str, type[SyftObject]] = {
            klass.__canonical_name__: klass for klass in action_object_types
        }

        action_object_pending_migration = self._find_klasses_pending_for_migration(
            context=context, object_types=action_object_types
        ).unwrap()
        result_dict: dict[type[SyftObject], list[SyftObject]] = defaultdict(list)
        action_stash = context.server.services.action.stash
        action_store_objects = action_stash.get_all(context.credentials).unwrap()

        for obj in action_store_objects:
            if get_all or type(obj) in action_object_pending_migration:
                klass = klass_by_canonical_name.get(obj.__canonical_name__, type(obj))
                result_dict[klass].append(obj)  # type: ignore
        return dict(result_dict)

    @as_result(SyftException)
    def _update_migrated_actionobjects(
        self, context: AuthedServiceContext, objects: list[SyftObject]
    ) -> str:
        action_store: ActionObjectStash = context.server.services.action.stash
        for obj in objects:
            action_store.set_or_update(
                uid=obj.id,
                credentials=context.credentials,
                syft_object=obj,
            ).unwrap()
        return "success"

    @service_method(
        path="migration.get_migration_data",
        name="get_migration_data",
        roles=ADMIN_ROLE_LEVEL,
    )
    def get_migration_data(self, context: AuthedServiceContext) -> MigrationData:
        store_objects = self._get_migration_objects(context, get_all=True).unwrap()
        action_objects = self._get_migration_actionobjects(
            context, get_all=True
        ).unwrap()
        blob_storage_objects = store_objects.pop(BlobStorageEntry, [])
        store_metadata = self._get_all_store_metadata(context).unwrap()
        return MigrationData(
            server_uid=context.server.id,
            signing_key=context.server.signing_key,
            syft_version=syft.__version__,
            default_pool_name=DEFAULT_WORKER_POOL_NAME,
            store_objects=store_objects,
            metadata=store_metadata,
            action_objects=action_objects,
            blob_storage_objects=blob_storage_objects,
        )

    @service_method(
        path="migration.apply_migration_data",
        name="apply_migration_data",
        roles=ADMIN_ROLE_LEVEL,
        unwrap_on_success=False,
    )
    def apply_migration_data(
        self,
        context: AuthedServiceContext,
        migration_data: MigrationData,
    ) -> SyftSuccess:
        # NOTE blob storage is migrated via client,
        # it needs access to both source and destination blob storages.
        if len(migration_data.blobs):
            raise SyftException(
                public_message="Blob storage migration is not supported by this endpoint, "
                "please use 'client.load_migration_data' instead."
            )

        created_objects = self._create_migrated_objects(
            context, migration_data.store_objects, skip_check_type=True
        ).unwrap()

        # migrate + apply store objects
        migrated_objects = self._migrate_objects(
            context,
            created_objects,
        ).unwrap()
        self._update_migrated_objects(context, migrated_objects).unwrap()

        # migrate+apply action objects
        migrated_actionobjects = self._migrate_objects(
            context, migration_data.action_objects
        ).unwrap()
        self._update_migrated_actionobjects(context, migrated_actionobjects).unwrap()

        # apply metadata
        self._update_store_metadata(context, migration_data.metadata).unwrap()
        return SyftSuccess(message="Migration completed successfully")

    @service_method(
        path="migration.reset_and_restore",
        name="reset_and_restore",
        roles=ADMIN_ROLE_LEVEL,
        unwrap_on_success=False,
    )
    def reset_and_restore(
        self,
        context: AuthedServiceContext,
        migration_data: MigrationData,
    ) -> SyftSuccess | SyftError:
        try:
            root_verify_key = context.server.verify_key
            context.server.db.init_tables(reset=True)
            context.credentials = root_verify_key
            self.apply_migration_data(context, migration_data)
        except Exception as e:
            return SyftError.from_exception(
                context=context,
                exc=e,
                include_traceback=True,
            )

        return SyftSuccess(message="Database reset successfully.")

    @service_method(
        path="migration._get_object",
        name="_get_object",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def _get_object(
        self, context: AuthedServiceContext, uid: UID, object_type: type
    ) -> Any:
        return (
            get_store_by_type(context, object_type)
            .get_by_uid(credentials=context.credentials, uid=uid)
            .unwrap()
        )

    @service_method(
        path="migration._update_object",
        name="_update_object",
        roles=ADMIN_ROLE_LEVEL,
    )
    def _update_object(self, context: AuthedServiceContext, object: Any) -> Any:
        return (
            get_store(context, object)
            .update(credentials=context.credentials, obj=object)
            .unwrap()
        )



# stdlib
from collections.abc import Callable
from io import BytesIO
from pathlib import Path
import sys
from typing import Any

# third party
from typing_extensions import Self
import yaml

# relative
from ...serde.deserialize import _deserialize
from ...serde.serializable import serializable
from ...serde.serialize import _serialize
from ...server.credentials import SyftSigningKey
from ...server.credentials import SyftVerifyKey
from ...store.db.stash import ObjectStash
from ...store.document_store_errors import NotFoundException
from ...types.blob_storage import BlobStorageEntry
from ...types.blob_storage import CreateBlobStorageEntry
from ...types.errors import SyftException
from ...types.result import as_result
from ...types.syft_migration import migrate
from ...types.syft_object import Context
from ...types.syft_object import SYFT_OBJECT_VERSION_1
from ...types.syft_object import SYFT_OBJECT_VERSION_2
from ...types.syft_object import SyftBaseObject
from ...types.syft_object import SyftObject
from ...types.syft_object_registry import SyftObjectRegistry
from ...types.transforms import make_set_default
from ...types.uid import UID
from ...util.util import prompt_warning_message
from ..response import SyftSuccess
from ..worker.utils import DEFAULT_WORKER_POOL_NAME
from ..worker.worker_image import SyftWorkerImage
from ..worker.worker_pool import SyftWorker
from ..worker.worker_pool import WorkerPool


@serializable()
class SyftObjectMigrationState(SyftObject):
    __canonical_name__ = "SyftObjectMigrationState"
    __version__ = SYFT_OBJECT_VERSION_1

    __attr_unique__ = ["canonical_name"]

    canonical_name: str
    current_version: int

    @property
    def latest_version(self) -> int | None:
        available_versions = SyftObjectRegistry.get_versions(
            canonical_name=self.canonical_name,
        )
        if not available_versions:
            return None

        return sorted(available_versions, reverse=True)[0]

    @property
    def supported_versions(self) -> list:
        return SyftObjectRegistry.get_versions(self.canonical_name)


@serializable(canonical_name="SyftMigrationStateSQLStash", version=1)
class SyftMigrationStateStash(ObjectStash[SyftObjectMigrationState]):
    @as_result(SyftException, NotFoundException)
    def get_by_name(
        self, canonical_name: str, credentials: SyftVerifyKey
    ) -> SyftObjectMigrationState:
        return self.get_one(
            credentials=credentials,
            filters={"canonical_name": canonical_name},
        ).unwrap()


@serializable()
class StoreMetadata(SyftBaseObject):
    __canonical_name__ = "StoreMetadata"
    __version__ = SYFT_OBJECT_VERSION_1

    object_type: type
    permissions: dict[UID, set[str]]
    storage_permissions: dict[UID, set[UID]]


@serializable()
class MigrationData(SyftObject):
    __canonical_name__ = "MigrationData"
    __version__ = SYFT_OBJECT_VERSION_2
    syft_version: str = ""
    default_pool_name: str = DEFAULT_WORKER_POOL_NAME
    server_uid: UID
    signing_key: SyftSigningKey
    store_objects: dict[type[SyftObject], list[SyftObject]]
    metadata: dict[type[SyftObject], StoreMetadata]
    action_objects: dict[type[SyftObject], list[SyftObject]]
    blob_storage_objects: list[SyftObject]
    blobs: dict[UID, Any] = {}

    __repr_attrs__ = [
        "server_uid",
        "root_verify_key",
        "num_objects",
        "num_action_objects",
        "includes_blobs",
    ]

    @property
    def root_verify_key(self) -> SyftVerifyKey:
        return self.signing_key.verify_key

    @property
    def num_objects(self) -> int:
        return sum(len(objs) for objs in self.store_objects.values())

    @property
    def num_action_objects(self) -> int:
        return sum(len(objs) for objs in self.action_objects.values())

    @property
    def includes_blobs(self) -> bool:
        blob_ids = [obj.id for obj in self.blob_storage_objects]
        return set(self.blobs.keys()) == set(blob_ids)

    @property
    def includes_custom_workerpools(self) -> bool:
        cname = WorkerPool.__canonical_name__
        worker_pools = None
        for k, v in self.store_objects.items():
            if k.__canonical_name__ == cname:
                worker_pools = v

        if worker_pools is None:
            return False

        custom_pools = [
            pool
            for pool in worker_pools
            if getattr(pool, "name", None) != self.default_pool_name
        ]
        return len(custom_pools) > 0

    def make_migration_config(self) -> dict[str, Any]:
        server_uid = self.server_uid.to_string()
        server_private_key = str(self.signing_key)
        migration_config = {
            "server": {
                "env": [
                    {"name": "SERVER_UID", "value": server_uid},
                    {"name": "SERVER_PRIVATE_KEY", "value": server_private_key},
                ]
            }
        }
        return migration_config

    @classmethod
    def from_file(self, path: str | Path) -> Self:
        path = Path(path)
        if not path.exists():
            raise SyftException(f"File {str(path)} does not exist.")

        with open(path, "rb") as f:
            res: SyftObject = _deserialize(f.read(), from_bytes=True)

        if not isinstance(res, MigrationData):
            latest_version = SyftObjectRegistry.get_latest_version(  # type: ignore[unreachable]
                MigrationData.__canonical_name__
            )
            print("Upgrading MigrationData object to latest version...")
            res = res.migrate_to(latest_version)

        return res

    def save(self, path: str | Path, yaml_path: str | Path) -> SyftSuccess:
        if not self.includes_blobs:
            proceed = prompt_warning_message(
                "You are saving migration data without blob storage data. "
                "This means that any existing blobs will be missing when you load this data."
                "\nTo include blobs, call `download_blobs()` before saving.",
                confirm=True,
            )
            if not proceed:
                raise SyftException(message="Migration data not saved.")

        path = Path(path)
        with open(path, "wb") as f:
            f.write(_serialize(self, to_bytes=True))

        yaml_path = Path(yaml_path)
        migration_config = self.make_migration_config()
        with open(yaml_path, "w") as f:
            yaml.dump(migration_config, f)

        return SyftSuccess(message=f"Migration data saved to {str(path)}.")

    def download_blobs(self) -> None:
        for obj in self.blob_storage_objects:
            blob = self.download_blob(obj.id)
            self.blobs[obj.id] = blob
        return None

    def download_blob(self, obj_id: str) -> Any:
        api = self._get_api()
        blob_retrieval = api.services.blob_storage.read(obj_id)
        return blob_retrieval.read()

    def migrate_and_upload_blobs(self) -> SyftSuccess:
        for obj in self.blob_storage_objects:
            self.migrate_and_upload_blob(obj)
        return SyftSuccess(message="All blobs uploaded successfully.")

    def migrate_and_upload_blob(self, obj: BlobStorageEntry) -> SyftSuccess:
        api = self._get_api()

        if obj.id not in self.blobs:
            raise SyftException(
                public_message=f"Blob {obj.id} not found in migration data."
            )
        data = self.blobs[obj.id]

        migrated_obj = obj.migrate_to(BlobStorageEntry.__version__, Context())
        serialized = _serialize(data, to_bytes=True)
        size = sys.getsizeof(serialized)
        blob_create = CreateBlobStorageEntry.from_blob_storage_entry(migrated_obj)
        blob_create.file_size = size
        blob_deposit_object = api.services.blob_storage.allocate_for_user(
            blob_create, migrated_obj.uploaded_by
        )
        return blob_deposit_object.write(BytesIO(serialized)).unwrap()

    def get_items_by_canonical_name(self, canonical_name: str) -> list[SyftObject]:
        for k, v in self.store_objects.items():
            if k.__canonical_name__ == canonical_name:
                return v

        for k, v in self.action_objects.items():
            if k.__canonical_name__ == canonical_name:
                return v
        return []

    def get_metadata_by_canonical_name(self, canonical_name: str) -> StoreMetadata:
        for k, v in self.metadata.items():
            if k.__canonical_name__ == canonical_name:
                return v
        return StoreMetadata(
            object_type=SyftObject, permissions={}, storage_permissions={}
        )

    def copy_without_workerpools(self) -> "MigrationData":
        items_to_exclude = [
            WorkerPool.__canonical_name__,
            SyftWorkerImage.__canonical_name__,
            SyftWorker.__canonical_name__,
        ]

        store_objects = {
            k: v
            for k, v in self.store_objects.items()
            if k.__canonical_name__ not in items_to_exclude
        }
        metadata = {
            k: v
            for k, v in self.metadata.items()
            if k.__canonical_name__ not in items_to_exclude
        }
        return self.__class__(
            server_uid=self.server_uid,
            signing_key=self.signing_key,
            store_objects=store_objects,
            metadata=metadata,
            action_objects=self.action_objects,
            blob_storage_objects=self.blob_storage_objects,
            blobs=self.blobs,
        )

    def copy_without_blobs(self) -> "MigrationData":
        # Create a shallow copy of the MigrationData instance, removing blob-related data
        # This is required for sending the MigrationData to the backend.
        copy_data = self.__class__(
            server_uid=self.server_uid,
            signing_key=self.signing_key,
            store_objects=self.store_objects.copy(),
            metadata=self.metadata.copy(),
            action_objects=self.action_objects.copy(),
            blob_storage_objects=[],
            blobs={},
        )
        return copy_data


@serializable()
class MigrationDataV1(SyftObject):
    __canonical_name__ = "MigrationData"
    __version__ = SYFT_OBJECT_VERSION_1

    server_uid: UID
    signing_key: SyftSigningKey
    store_objects: dict[type[SyftObject], list[SyftObject]]
    metadata: dict[type[SyftObject], StoreMetadata]
    action_objects: dict[type[SyftObject], list[SyftObject]]
    blob_storage_objects: list[SyftObject]
    blobs: dict[UID, Any] = {}


@migrate(MigrationDataV1, MigrationData)
def migrate_migrationdata_v1_to_v2() -> list[Callable]:
    return [
        make_set_default("default_pool_name", DEFAULT_WORKER_POOL_NAME),
        make_set_default("syft_version", ""),
    ]



# stdlib
from collections.abc import Callable
import logging
import os
from typing import Any
from typing import TYPE_CHECKING
from typing import TypeVar

# relative
from .. import __version__
from .util import str_to_bool

__all__ = [
    "TRACING_ENABLED",
    "instrument",
    "instrument_fastapi",
    "instrument_botocore",
]

TRACING_ENABLED = str_to_bool(os.environ.get("TRACING", "False"))
logger = logging.getLogger(__name__)

T = TypeVar("T", bound=Callable | type)


def no_instrument(__func_or_class: T | None = None, /, *args: Any, **kwargs: Any) -> T:
    def noop_wrapper(__func_or_class: T) -> T:
        return __func_or_class

    if __func_or_class is None:
        return noop_wrapper  # type: ignore
    else:
        return __func_or_class


def setup_instrumenter() -> Any:
    if not TRACING_ENABLED:
        return no_instrument

    try:
        # third party
        from opentelemetry import trace
        from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import (
            OTLPSpanExporter,
        )
        from opentelemetry.sdk.resources import OTELResourceDetector
        from opentelemetry.sdk.resources import ProcessResourceDetector
        from opentelemetry.sdk.resources import Resource
        from opentelemetry.sdk.trace import TracerProvider
        from opentelemetry.sdk.trace.export import BatchSpanProcessor

        # relative
        from .trace_decorator import instrument

        # create a resource
        resource = Resource({"syft.version": __version__})
        resource = resource.merge(OTELResourceDetector().detect())
        resource = resource.merge(ProcessResourceDetector().detect())
        logger.debug(f"OTEL resource : {resource.__dict__}")

        # create a trace provider from the resource
        provider = TracerProvider(resource=resource)

        # create a span processor
        otlp_exporter = OTLPSpanExporter()
        span_processor = BatchSpanProcessor(otlp_exporter)
        provider.add_span_processor(span_processor)

        # set the global trace provider
        trace.set_tracer_provider(provider)

        logger.info("Added TracerProvider with BatchSpanProcessor")
        return instrument
    except Exception as e:
        logger.error("Failed to import opentelemetry", exc_info=e)
        return no_instrument


def instrument_fastapi(app: Any) -> None:
    if not TRACING_ENABLED:
        return

    try:
        # third party
        from opentelemetry.instrumentation.fastapi import FastAPIInstrumentor

        FastAPIInstrumentor().instrument_app(app)
        logger.info("Added OTEL FastAPIInstrumentor")
    except Exception as e:
        logger.error(f"Failed to load FastAPIInstrumentor. {e}")


def instrument_botocore() -> None:
    if not TRACING_ENABLED:
        return

    try:
        # third party
        from opentelemetry.instrumentation.botocore import BotocoreInstrumentor

        BotocoreInstrumentor().instrument()
        logger.info("Added OTEL BotocoreInstrumentor")
    except Exception as e:
        logger.error(f"Failed to load BotocoreInstrumentor. {e}")


def instrument_threads() -> None:
    if not TRACING_ENABLED:
        return

    try:
        # third party
        from opentelemetry.instrumentation.threading import ThreadingInstrumentor

        ThreadingInstrumentor().instrument()
        logger.info("Added OTEL ThreadingInstrumentor")
    except Exception as e:
        logger.error(f"Failed to load ThreadingInstrumentor. {e}")


def instrument_sqlalchemny() -> None:
    if not TRACING_ENABLED:
        return

    try:
        # third party
        from opentelemetry.instrumentation.sqlalchemy import SQLAlchemyInstrumentor

        SQLAlchemyInstrumentor().instrument(enable_commenter=True, commenter_options={})
        logger.info("Added OTEL SQLAlchemyInstrumentor")
    except Exception as e:
        logger.error(f"Failed to load SQLAlchemyInstrumentor. {e}")


if TYPE_CHECKING:
    # To let static type checker know the returntype of instrument decorators
    instrument = no_instrument
else:
    instrument = setup_instrumenter()



# stdlib
import json
import subprocess  # nosec
import sys

# relative
from ..service.worker.utils import DEFAULT_WORKER_POOL_NAME
from ..types.uid import UID
from ..util.util import get_env
from ..util.util import str_to_bool

SERVER_PRIVATE_KEY = "SERVER_PRIVATE_KEY"
SERVER_UID = "SERVER_UID"
SERVER_TYPE = "SERVER_TYPE"
SERVER_NAME = "SERVER_NAME"
SERVER_SIDE_TYPE = "SERVER_SIDE_TYPE"

DEFAULT_ROOT_EMAIL = "DEFAULT_ROOT_EMAIL"
DEFAULT_ROOT_USERNAME = "DEFAULT_ROOT_USERNAME"
DEFAULT_ROOT_PASSWORD = "DEFAULT_ROOT_PASSWORD"  # nosec


def get_private_key_env() -> str | None:
    return get_env(SERVER_PRIVATE_KEY)


def get_server_type() -> str | None:
    return get_env(SERVER_TYPE, "datasite")


def get_server_name() -> str | None:
    return get_env(SERVER_NAME, None)


def get_server_side_type() -> str | None:
    return get_env(SERVER_SIDE_TYPE, "high")


def get_server_uid_env() -> str | None:
    return get_env(SERVER_UID)


def get_default_root_email() -> str | None:
    return get_env(DEFAULT_ROOT_EMAIL, "info@openmined.org")


def get_default_root_username() -> str | None:
    return get_env(DEFAULT_ROOT_USERNAME, "Jane Doe")


def get_default_root_password() -> str | None:
    return get_env(DEFAULT_ROOT_PASSWORD, "changethis")  # nosec


def get_enable_warnings() -> bool:
    return str_to_bool(get_env("ENABLE_WARNINGS", "False"))


def get_container_host() -> str | None:
    return get_env("CONTAINER_HOST")


def get_default_worker_image() -> str | None:
    return get_env("DEFAULT_WORKER_POOL_IMAGE")


def get_default_worker_pool_name() -> str | None:
    return get_env("DEFAULT_WORKER_POOL_NAME", DEFAULT_WORKER_POOL_NAME)


def get_default_bucket_name() -> str:
    env = get_env("DEFAULT_BUCKET_NAME")
    server_id = get_server_uid_env() or "syft-bucket"
    return env or server_id or "syft-bucket"


def get_default_worker_pool_pod_annotations() -> dict[str, str] | None:
    annotations = get_env("DEFAULT_WORKER_POOL_POD_ANNOTATIONS", "null")
    return json.loads(annotations)


def get_default_worker_pool_pod_labels() -> dict[str, str] | None:
    labels = get_env("DEFAULT_WORKER_POOL_POD_LABELS", "null")
    return json.loads(labels)


def in_kubernetes() -> bool:
    return get_container_host() == "k8s"


def get_venv_packages() -> str:
    try:
        # subprocess call is safe because it uses a fully qualified path and fixed arguments
        result = subprocess.run(
            [sys.executable, "-m", "pip", "list", "--format=freeze"],  # nosec
            capture_output=True,
            check=True,
            text=True,
        )
        return result.stdout
    except subprocess.CalledProcessError as e:
        return f"An error occurred: {e.stderr}"


def get_syft_worker() -> bool:
    return str_to_bool(get_env("SYFT_WORKER", "false"))


def get_k8s_pod_name() -> str | None:
    return get_env("K8S_POD_NAME")


def get_syft_worker_uid() -> str | None:
    is_worker = get_syft_worker()
    pod_name = get_k8s_pod_name()
    uid = get_env("SYFT_WORKER_UID")
    # if uid is empty is a K8S worker, generate a uid from the pod name
    if (not uid) and is_worker and pod_name:
        uid = str(UID.with_seed(pod_name))
    return uid
